/* Copyright 2017, 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* FIXME:

   radius==0 in n=0, follow pattern of floor (diameter+1)/2 ?

   names _and_count just as _info ?

   minimal domsets on forest

   diameter v1,v2 make v1 the deeper maybe, or give depths or distances to
   common ancestor
*/



/* This is some extra functions for use with the nauty graph theory library,

       http://cs.anu.edu.au/~bdm/nauty
       http://pallini.di.uniroma1.it

   See comments with each function for usage.
   nautyextra.h summarizes what's available.
   nautyextra.h can be used from C++ too.
   See the test directory for some self-tests.

   Only source compatibility for now, and struct field order not promised
   (might add more fields or re-arrange).  The suggestion is to static link
   nautyextra.c into a program.

   "vpar" is per gentreeg.c OUTPROC.  It's an array of int vpar[1..n]
   representing a rooted labelled tree or forest of n vertices.  Each
   vpar[v] is the parent vertex number of v.  Most vpar invariants
   (diameter, domnum, etc) are calculated in linear time, by 2 or 3 passes
   over the array.

   There's no configuration as such, but some compile time options:

   -DMAXN=123
      A fixed maximum number of vertices, like in nauty.
      The default is dynamic allocation, limited only by memory.
      Most functions have static allocation (on the stack) if compiled with
      MAXN.  The limit applies to both inputs and outputs.

      graph and sparsegraph operations are on vertices 0 to n-1 inclusive so
      in the code their arrays are sized foo[MAXN].  vpar operations are on
      vertices 1 to n so they are foo[MAXN+1].

      A few naive set-iterating functions are further limited to n<=WORDSIZE
      vertices (32 or 64), since iterating 2^32 is usually too much already
      so don't need multi-word iterating code.

   -DNAUTYEXTRA_WANT_ASSERTS=1
      Enable assert() and similar consistency checking.  The default is no
      checking.

      The asserts are sprayed liberally and are a mixture of input
      checking, and self-checking on the code and the algorithms.
      Most are a small to medium slow-down.  The suggestion is to have them
      on for test program or when debugging, and off for full speed run.

   -DNAUTYEXTRA_WANT_OVERFLOW_DETECTION=0
      Disable overflow checking in some of the count or size operations.
      The default is to have such checks, primarily so that garbage isn't
      quietly returned.

      The multiply overflow detection is currently a medium slow-down for
      things like indsets counting, so disable if confident about the range
      of the results (and some headroom above sometimes).

   -DNAUTYEXTRA_TYPE_ULL="unsigned"
      An integer type to use for various counts.  The default is "unsigned
      long long", but if confident something smaller is enough then reducing
      it might be a speedup.  This type appears in nautyextra.h too so
      remember to compile both this file and the mainline with the same.

   -DWANT_DEBUG=1
      Enable a lot of printf()s of function progress etc.  These are
      intended only for debugging the code itself.  The output is quite
      verbose and rather specific to what has needed tracing.
*/


/* assert()s and similar check code disabled by default.
   Compile with -DNAUTYEXTRA_WANT_ASSERTS=1 to have asserts.
*/
#if ! NAUTYEXTRA_WANT_ASSERTS
#if ! defined(NDEBUG)  /* if NDEBUG not already set */
#define NDEBUG
#endif
#endif
#include <assert.h>

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <nauty/nauty.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"

/* Change WANT_DEBUG to 1 for mucho debug prints. */
#ifndef WANT_DEBUG
#define WANT_DEBUG  0
#endif
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

/* Default is to detect overflows in various counts, Wiener index, etc.
   Compile with -DNAUTYEXTRA_WANT_OVERFLOW_DETECTION=0 to disable.
*/
#if ! defined(NAUTYEXTRA_WANT_OVERFLOW_DETECTION)
#define NAUTYEXTRA_WANT_OVERFLOW_DETECTION 1
#endif

/* ------------------------------------------------------------------------- */
/* Macros Used Here */

/* ss is a setword variable "setword ss".
   n is an int number of bits to have in ss.
   Loop over all sets of n bits in ss.
   Evaluate "body" with successive ss.
   body can goto or return etc to abort the loop.
*/
#define NAUTYEXTRA_SET1_LOOP(ss,n,body)                 \
  do {                                                  \
    assert (n >= 1);                                    \
    if (! (SETWORDSNEEDED(n) == 1)) abort();            \
    setword __NAUTYEXTRA_SET_LOOP__inc = BITT[n-1];     \
    (ss) = 0;                                           \
    do {                                                \
      body;                                             \
    } while ((ss) += __NAUTYEXTRA_SET_LOOP__inc);       \
  } while (0)

static void
INTERNAL_overflow (void)
{
  fprintf(stderr, "nautyextra integer overflow\n");
  abort();
}

/* dst,x,y are NAUTYEXTRA_TYPE_ULL variables.
   Set dst = x+y, with abort on overflow. */
#define ADD_NO_OVERFLOW_ULL(dst, x,y)                                   \
  do {                                                                  \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__x = (x);                 \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__y = (y);                 \
    NAUTYEXTRA_TYPE_ULL __ADD_NO_OVERFLOW_ULL__ret                      \
      = __ADD_NO_OVERFLOW_ULL__x + __ADD_NO_OVERFLOW_ULL__y;            \
    (dst) = __ADD_NO_OVERFLOW_ULL__ret;                                 \
    if (NAUTYEXTRA_WANT_OVERFLOW_DETECTION) {                           \
      if (__ADD_NO_OVERFLOW_ULL__ret < __ADD_NO_OVERFLOW_ULL__x)        \
        INTERNAL_overflow();                                          \
    }                                                                   \
  } while (0)

/* dst,x,y are NAUTYEXTRA_TYPE_ULL variables.
   Set dst = x*y, with abort on overflow.
   FIXME: Division to detect overflow will be a bit slow.
   Maybe a mult in floating point would be better.
*/
#define MUL_NO_OVERFLOW_ULL(dst, x,y)                                   \
  do {                                                                  \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__x = (x);                 \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__y = (y);                 \
    NAUTYEXTRA_TYPE_ULL __MUL_NO_OVERFLOW_ULL__ret                      \
      = __MUL_NO_OVERFLOW_ULL__x * __MUL_NO_OVERFLOW_ULL__y;            \
    (dst) = __MUL_NO_OVERFLOW_ULL__ret;                                 \
    if (NAUTYEXTRA_WANT_OVERFLOW_DETECTION) {                           \
      if (__MUL_NO_OVERFLOW_ULL__x                                      \
          && __MUL_NO_OVERFLOW_ULL__ret / __MUL_NO_OVERFLOW_ULL__x      \
          != __MUL_NO_OVERFLOW_ULL__y)                                  \
        INTERNAL_overflow();                                          \
    }                                                                   \
  } while (0)

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))

/* number of bits in a NAUTYEXTRA_TYPE_ULL */
#define ULL_BIT  (CHAR_BIT*sizeof(NAUTYEXTRA_TYPE_ULL))

/* ------------------------------------------------------------------------- */
/* Static Workspace */

#if ! MAXN

DYNALLSTAT(char,INTERNAL_workspace,INTERNAL_workspace_sz);
#define WORKSPACE(type,var,size,name)                          \
  DYNALLOC1(char,                                              \
            INTERNAL_workspace,INTERNAL_workspace_sz,          \
            (size)*sizeof(type), name);                        \
  type *var = (type*) INTERNAL_workspace;

#endif

/* ------------------------------------------------------------------------- */
/* Diagnostics */

/* Print various parts of sg.
   Cf nauty putgraph_sg(), but here showing more of the internals.
   The exact output form is unspecified, but meant for human readers.
*/
void
nautyextra_print_sg (const sparsegraph *sg)
{
  int i, num;

  printf ("  nv  %d  vlen=%d (%p), dlen=%d (%p)\n",
          sg->nv,
          sg->vlen,sg->v,
          sg->dlen,sg->d);
  printf ("  nde %d elen=%d  (%p)\n", sg->nde, sg->elen, sg->e);

  printf ("  d outdegrees ");
  for (i = 0; i < sg->nv; i++) {
    if (i > 100) {
      printf(" ...");
      break;
    }
    printf (" %d", sg->d[i]);
  }
  printf ("\n");

  printf ("  v pos in e   ");
  for (i = 0; i < sg->nv; i++) {
    if (i > 100) {
      printf(" ...");
      break;
    }
    printf (" %d", sg->v[i]);
  }
  printf ("\n");

  printf ("  e edges    ");
  num = sg->nde;
  if (num > sg->elen) { num = sg->elen; }
  for (i = 0; i < num; i++) {
    if (i > 100) {
      printf(" ...");
      break;
    }
    printf (" %d", sg->e[i]);
  }
  printf ("\n");
}

/* Print the parent vertex list vpar[1..n].
   The exact output form is unspecified, but meant for human readers.
*/
void
nautyextra_print_vpar (const int *vpar, int n)
{
  int v;
  assert (n >= 0);

  printf("n=%d ",n);
  const char *join = "";
  for (v=1; v<=n; v++) {
    printf("%s%d", join, vpar[v]);
    join = ",";
  }
  putchar('\n');
}

/* Return a \0 terminated string of characters "0" or "1" for the first n
   bits of ss, being elements 0 to n-1.

   The string is a static allocation, good until the next call here.
   It has TLS_ATTR so calls in other threads are ok.
*/
char *
nautyextra_set_str(const set *ss, int n)
{
  int i;
#if MAXN
  static TLS_ATTR char str[MAXN+1];  /* with \0 terminator */
  assert(n <= MAXN);
#else
  DYNALLSTAT(char,str,str_sz);
  DYNALLOC1(char,str,str_sz,n+1,"nautyextra_set_str");
#endif
  assert (n >= 0);

  for (i = 0; i < n; i++)
    str[i] = (ISELEMENT(ss,i) ? '1' : '0');
  str[n] = '\0';
  return str;
}


/* ------------------------------------------------------------------------- */
/* Validating */

/* Validate the contents of graph g,m,n.
   If good then return TRUE.  If bad then print messages and abort().
   The return value allows for example assert(nautyextra_validate(g,m,n))
   which runs only when debugging etc.

   The only check is to see that the m rows have bits set only in the range
   0 to n-1.  A few functions, and believe including some of nauty() itself,
   depend on g having bits only in that range and 0s in the rest of the last
   setword and possible additional setwords.
*/
int
nautyextra_validate (const graph *g, int m, int n)
{
  DEBUG(printf("nautyextra_validate() g=%p m=%d n=%d\n", g,m,n));

  const char *message = "nautyextra_validate() bad:";

  /* allow g==NULL when n==0 since DYNALLOC2() for instance leaves g=NULL
     when asked for a 0 vertex graph */
  if (n && g == NULL) {
    fprintf(stderr, "%s g==NULL\n", message);
    abort();
  }
  int m_from_n = SETWORDSNEEDED(n);
  if (m < m_from_n) {
    DEBUG(printf("graph n=%d,m=%d but need m>=%d for that n\n",
                 n,m, m_from_n));
    abort();
  }
  int pad = m_from_n*WORDSIZE - n;
  assert (pad >= 0 && pad <= WORDSIZE-1);
  setword pad_mask = (pad ? BITMASK(WORDSIZE-1-pad) : 0);
  DEBUG(printf("  m_from_n %d  pad %d  pad_mask %X\n",
               m_from_n, pad, pad_mask));
  int v;
  for (v=0; v<n; v++,g+=m) {
    int s = g[m_from_n-1];
    if (s & pad_mask) {
      fprintf(stderr, "%s graph m=%d,n=%d extra bits in vertex %d row, partial setword 0x",
              message, m,n, v);
      fprintf(stderr, SETWORD_FORMAT, s);
      fprintf(stderr, "\n");
      abort();
    }
    int i;
    for (i = m_from_n; i<m; i++) {
      s = g[i];
      if (s) {
        fprintf(stderr, "%s graph m=%d n=%d extra bits in v=%d row whole extra setword 0x",
                message, m,n, v);
        fprintf(stderr, SETWORD_FORMAT, s);
        fprintf(stderr, " at %d\n", i);
        abort();
      }
    }
  }

  DEBUG(printf(" ok\n"));
  return TRUE;
}

/* Validate the contents of sg, to guard against indices outside nv and elen
   ranges, etc.
   If good then return TRUE.  If bad then print messages and abort().
   The return value allows for example assert(nautyextra_validate_sg(sg))
   which runs only when debugging etc.

   Overlapping regions of sg->e are reckoned bad.  In principle it could
   work for read-only of something with some common structure, but when
   making a general graph it's almost certainly a mistake.
 */
int
nautyextra_validate_sg (const sparsegraph *sg)
{
  const char *message = "nautyextra_validate_sg() bad:";
  if (sg == NULL) {
    fprintf(stderr,"%s sg==NULL\n", message);
    abort();
  }

  const int bad_limit = 10;
  int bad = 0;
  int nv = sg->nv;
  int nde = sg->nde;
  size_t elen = sg->elen;
  DEBUG(printf("nautyextra_validate_sg() nv=%d dlen=%d vlen=%d  nde=%d elen=%d\n",
               nv, sg->dlen, sg->vlen, nde, elen));

  if (! (nv >= 0)) {
    fprintf (stderr,"%s nv negative\n", message);
    abort();
  }
#if MAXN
  if (! (nv <= MAXN)) {
    fprintf (stderr,"%s nv bigger than MAXN\n", message);
    abort();
  }
#endif

  if (sg->dlen < nv) {
    fprintf (stderr,"%s dlen=%d too small for nv=%d\n",
             message, sg->dlen, nv);
    message = "";
    bad++;
  }
  if (sg->vlen < nv) {
    fprintf (stderr,"%s vlen=%d too small for nv=%d\n",
             message, sg->vlen, nv);
    message = "";
    bad++;
  }
  if (bad) abort();

  if (nde < 0) {
    fprintf (stderr,"%s nde=%d negative\n", message, nde);
    abort();
  }
  if (elen < nde) {
    fprintf (stderr,"%s elen=%d too small for nde=%d\n",
             message, elen, nde);
    message = "";
    bad++;
  }

  const int    *dd = sg->d;
  const size_t *vv = sg->v;
  const int    *ee = sg->e;

  /* when vertices !=0 must have dd!=NULL and vv!=NULL */
  if (nv && dd==NULL) {
    fprintf (stderr,"%s nv=%d but d=NULL\n", message, nv);
    message = "";
    bad++;
  }
  if (nv && vv==NULL) {
    fprintf (stderr,"%s nv=%d but v=NULL\n", message, nv);
    message = "";
    bad++;
  }
  if (bad) abort();

  /* cast since SETWORDSNEEDED is for signed,
     unsigned 0 comes out as 0x8000... */
  int m = SETWORDSNEEDED((int)elen);

  assert (m >= 0);
  DEBUG(printf(" e seen set elen=%d 0x%X m=%d 0x%X\n", elen, elen, m, m));
  DYNALLSTAT(set,seen,seen_sz);
  DYNALLOC1(set,seen,seen_sz,m,"nautyextra_validate_sg");

  /* EMPTYSET doesn't like seen==NULL which occurs when m==0 */
  if (m) EMPTYSET(seen,m);

  int total_degree = 0;
  int v;
  for (v = 0; v < nv; v++) {
    int degree = dd[v];
    total_degree += degree;
    DEBUG(printf("v=%d outdegree %d\n", v, degree));
    if (degree < 0) {
      fprintf (stderr,"%s vertex %d degree=%d negative\n",
               message, v, degree);
      message = "";
      if (++bad > bad_limit) {
      TOO_MUCH_BAD:
        fprintf (stderr," stopped checking ...\n");
        abort();
      }
    }

    if (degree && ee==NULL) {       /* when any edge must have ee!=NULL */
      fprintf (stderr,"%s vertex %d degree=%d arcs but ee=NULL\n",
               message, v, degree);
      bad++;
      abort();
    }

    int epos = vv[v];
    DEBUG(printf("  vertex %d v epos %d\n", v, epos));
    if (! (epos >= 0 && epos <= elen)) {
      fprintf (stderr,"%s vertex %d epos v[%d]=%d out of range elen=%d\n",
               message, v, v, epos, elen);
      message = "";
      if (degree == 0)
        fprintf (stderr,"  (though degree==0 means not actually used)\n");
      if (++bad > bad_limit) goto TOO_MUCH_BAD;
    }

    int eafter = elen - epos;  /* space from epos to end of e */
    if (degree > eafter) {
      fprintf (stderr,"%s vertex %d epos v[%d]=%d degree=%d > space %d to elen=%d\n",
               message, v, v, epos, degree, eafter, elen);
      message = "";
      if (++bad > bad_limit) goto TOO_MUCH_BAD;
    }

    if (degree) {
      int i;
      for (i = 0; i < degree; i++) {
        int i_epos = epos + i;
        int to = ee[epos + i];
        DEBUG(printf("    edge i=%d %d to %d at epos=%d\n", i,v,to, i_epos));

        if (! (0 <= to && to < nv)) {
          fprintf (stderr,"%s vertex %d arc i=%d destination %d out of range nv=%d\n",
                   message, v, i, to, nv);
          message = "";
          if (++bad > bad_limit) goto TOO_MUCH_BAD;
        }

        if (ISELEMENT(seen,i_epos)) {
          fprintf (stderr,"%s vertex %d epos v[%d]=%d edge i=%d overlapping e use\n",
                   message, v, v, epos, i);
                   /* nautyextra_set_str(seen,elen)); */
          message = "";
          if (++bad > bad_limit) goto TOO_MUCH_BAD;
        }
        ADDELEMENT(seen,i_epos);
      }
    }
  }

  if (nde != total_degree) {
    fprintf (stderr,"%s nde=%d but total d degrees %d\n",
             message, nde, total_degree);
    message = "";
    bad++;
  }

  if (bad) abort();

  DEBUG(printf(" ok\n"));
  return TRUE;
}

static int
internal_validate_vpar (const int *vpar, int n, int want_abort)
{
  const char *message = "nautyextra_validate_vpar() ";
#if MAXN
  int seen[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  DYNALLSTAT(int,seen,seen_sz);
  DYNALLOC1(int,seen,seen_sz,n+1,"nautyextra_validate_vpar");
#endif
  memset(seen, '\0', (n+1)*sizeof(seen[0]));

  int bad = 0;
  int i;
  for (i = 1; i <= n; i++) {
    int v = i;
    do {
      int p = vpar[v];
      if (! (p >= 0 && p <= n)) {
        if (want_abort) {
          fputs(message,stderr); message = "";
          fprintf (stderr,"v=%d parent vpar[v]=%d out of range n=%d\n", v, p, n);
        }
        if (++bad > 10 && want_abort) abort();
        break;
      }
      v = p;
      if (seen[v]) {  /* revisit */
        if (seen[v] == i) {
          if (want_abort) {
            fputs(message,stderr); message = "";
            fprintf (stderr,"i=%d v=%d cycle\n", i, v);
          }
          if (++bad > 10 && want_abort) abort();
        }
        break;  /* revisit is to something previously found acyclic, good */
      }
      seen[v] = i;
    } while (v);
  }
  if (bad && want_abort) abort();
  return bad==0;
}

/* Check vpar[1..n] values.
   Return TRUE if good.   Print messages and abort() if not.

   The return means the check can be used as for example
       assert(nautyextra_validate_vpar(vpar,n))
   to have a check only for debugging.
*/
int
nautyextra_validate_vpar (const int *vpar, int n)
{
  return internal_validate_vpar(vpar,n, 1); /* want abort */
}


/* ------------------------------------------------------------------------- */
/* Directedness */

/* Return TRUE if g,m,n is undirected, meaning every u->v also has v->u.
   Return FALSE if not.
 */
int
nautyextra_is_undirected (const graph *g, int m, int n)
{
  const set *from_row = GRAPHROW(g,0,m);
  int from;
  for (from = 0; from < n; from++, from_row += m) {
    int to;
    for (to = -1; (to = nextelement((set*)from_row,m,to)) >= 0; )
      if (! ISELEMENT(GRAPHROW(g,to,m),from))
        return FALSE;
  }
  return TRUE;
}

/* Return TRUE if sg is undirected, meaning every arc u->v also has v->u.
   Return FALSE if not.
 */
int
nautyextra_is_undirected_sg (const sparsegraph *sg)
{
  int           n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;

  int from;
  for (from = 0; from < n; from++) {
    const int *from_eptr = ee + vv[from];
    int from_degree = dd[from];
    int from_i;
    for (from_i = 0; from_i < from_degree; from_i++) {
      int to             = *from_eptr++;
      const int *to_eptr = ee + vv[to];
      int to_degree      = dd[to];
      int to_i;
      for (to_i = 0; to_i < to_degree; to_i++) {
        int back = *to_eptr++;
        if (back == from) goto FOUND_BACK;
      }
      return FALSE;
    FOUND_BACK:
      ;
    }
  }
  return TRUE;
}



/* ------------------------------------------------------------------------- */
/* Sparsegraph Construction */

/* nautyextra_ensure_vertex_sg() ensures sg includes vertex v, by increasing
   sg->nv and the arrays as necessary.
   Any new vertices added have no edges to or from.
   v = -1 is allowed.  This is interpreted as desiring at least n=0 vertices,
   which is always true.
 */
void
nautyextra_ensure_vertex_sg (sparsegraph *sg, int v)
{
  assert (v >= -1);
  int new_nv = v+1;
  int nv = sg->nv;
  if (nv >= new_nv) return;
  sg->nv = new_nv;

  DEBUG(printf("nautyextra_ensure_vertex_sg() grow for v=%d\n", v));

  DYNREALLOC(size_t,sg->v,sg->vlen,new_nv, "nautyextra_ensure_vertex_sg");
  DYNREALLOC(int,   sg->d,sg->dlen,new_nv, "nautyextra_ensure_vertex_sg");

  size_t extralen = new_nv - nv;
  memset (sg->d + nv, '\0', extralen*sizeof(sg->d[0]));  /* out degree */
  memset (sg->v + nv, '\0', extralen*sizeof(sg->v[0]));  /* edges pos */
}

/* nautyextra_add_arc_sg() adds a new arc u-->v into sg.
   The number of vertices in sg is increased if necessary to ensure u and v
   exist.  Self-loops and multi-arcs can be added by this function.

   The arc is added by growing the sg->e array.
   If u has existing arcs then the new one is appended to their present
   location in sg->e and the rest of sg->e moved up.
   If u has no existing edges then it is put at the end of the present elen.

   If u has existing arcs then there's no attempt to identify a hole in
   sg->e after those arcs (which would make move and grow unnecessary).
   If u doesn't have existing arcs then there's no attempt to find a hole
   somewhere in sg->e which could be used (instead of extending end of e).
   These are both on the basis that when constructing there usually won't be
   any holes.  (After many big deletes a separate compaction could be done.)

   This shifting and reallocing sg->e is not particularly efficient and in
   the worst case can cause O(E^2) time making a graph.  But constructing
   like this is convenient before trying something more efficient setting sg
   parts directly.  Direct filling is not too difficult, but generally
   requires knowing in advance the out-degree of each vertex (or an upper
   bound on that) so as to allocate sg->e space to vertices.
*/
void
nautyextra_add_arc_sg (sparsegraph *sg, int u, int v)
{
  nautyextra_ensure_vertex_sg(sg, MAX(u,v));
  DEBUG(printf("nautyextra_add_arc_sg(%p)  %d %d\n", sg, u, v));
  DEBUG(printf("  nv=%d elen=%d(%d)\n",
               sg->nv, sg->elen,sg->elen));

  int nv  = sg->nv;
  assert (0 <= u && u < nv);
  assert (0 <= v && v < nv);

  size_t old_elen = sg->elen;
  size_t new_elen = old_elen + 1;
  DYNREALLOC(int, sg->e,sg->elen,new_elen, "nautyextra_add_arc_sg");

  sg->nde++;

  int    *ee = sg->e;
  size_t *vv = sg->v;
  int epos = vv[u];
  assert (0 <= epos && epos <= old_elen);

  int *dd = sg->d;
  int old_degree = dd[u]++;
  if (old_degree) {  /* u outdegree */

    /* Existing arcs at u, leave its e unchanged and move up everything after.
       The size after is old_elen - epos.
    */
    epos += old_degree;        /* after u */
    assert (epos <= old_elen);
    int *eptr = ee + epos;
    int move_len = old_elen - epos;
    DEBUG(printf("  move %d to %d length %d\n",
                 eptr+1-ee, eptr-ee, move_len));
    memmove (eptr+1, eptr, move_len*sizeof(ee[0]));

    DEBUG(printf("  eptr at %d set %d\n", eptr-ee, v));
    *eptr = v;   /* destination */

    /* move up later epos positions */
    int i;
    for (i = 0; i < nv; i++) {
      if (vv[i] >= epos) {
        DEBUG(printf("  shift up %d\n", i));
        assert (i != u);
        vv[i]++;
      }
    }

  } else {
    /* No other arcs at u, put at the end of e. */
    vv[u] = old_elen;
    ee[old_elen] = v;  /* destination */
  }

  DEBUG(nautyextra_print_sg(sg));
}

/* Add undirected edge u--v into sg, meaning arcs u-->v and u<--v.
   The number of vertices in sg is increased if necessary to ensure u and v
   exist.  See nautyextra_add_arc_sg() for details.
*/
void
nautyextra_add_edge_sg (sparsegraph *sg, int u, int v)
{
  nautyextra_add_arc_sg(sg,u,v);
  nautyextra_add_arc_sg(sg,v,u);
}


/* ------------------------------------------------------------------------- */
/* Isomorphic Graphs */

/* Return TRUE if g1,m,n and g2,m,n are the same contents.
   This is simply a memcmp() checking for equality of content.
   Similar to nauty aresame_sg() which is for sparsegraphs.
*/
int
nautyextra_aresame(const graph *g1, const graph *g2, int m, int n)
{
  return memcmp(g1, g2, m*n*sizeof(g1[0])) == 0;
}


/* ------------------------------------------------------------------------- */
/* Claws */

/* Return binomial coefficient (n,m). */
NAUTYEXTRA_TYPE_ULL
nautyextra_binomial (int n, int m)
{
  DEBUG(printf("nautyextra_binomial() %d %d\n", n,m));
  /* assert (n >= 0); */  /* no, allow n<0 to return 0 */
  if (! (0 <= m && m <= n)) return 0;

  if (m > (n>>1)) m = n - m;
  NAUTYEXTRA_TYPE_ULL ret = 1;
  int i;
  for (i = 1; i <= m; i++) {
    MUL_NO_OVERFLOW_ULL(ret, ret, n-i+1);
    assert ((ret % i) == 0);
    ret /= i;
  }
  return ret;
}

/* Return the number of induced claws in tree vpar[1..n].
   This is sum binomial(degree,3) for each vertex of degree >= 3.

   The return is type unsigned long long since the number of claws for a
   star goes as n^3/6.  A star of 2957 vertices exceeds 2^32.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_claw_count_vpar (const int *vpar, int n)
{
  /* GP-DEFINE  \\ maximum number of claws in a tree is for star-n
     GP-DEFINE  star_claw_count(n) = binomial(n-1,3);
     GP-Test  star_claw_count(4) == 1
     GP-Test  star_claw_count(2955) == 4291795704
     GP-Test  star_claw_count(2956) == 4296157285
     GP-Test  star_claw_count(2955) < 2^32
     GP-Test  star_claw_count(2956) > 2^32
     for(n=1,10000,if(star_claw_count(n)>2^32,print(n);break()));
  */

  DEBUG(printf("nautyextra_claw_count_vpar()\n"));
  int v;
#if MAXN
  int degrees[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,degrees,n+1,"nautyextra_claw_count_vpar");
#endif
  assert (n >= 0);

  nautyextra_degrees_array_vpar(vpar,n,degrees);
  int count = 0;
  for (v = 1; v <= n; v++) {
    int d = degrees[v];
    if (d >= 3)
      ADD_NO_OVERFLOW_ULL(count, count, nautyextra_binomial(d,3));
  }
  return count;
}

/* Return TRUE if tree vpar[1..n] is claw-free.
   Return FALSE if not (it has a claw).
   Claw-free means no vertex of degree >= 3, which means only path or forest
   of paths.
*/
int
nautyextra_is_clawfree_vpar (const int *vpar, int n)
{
  assert (n >= 0);
#if MAXN
  int degrees[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,degrees,n+1,"nautyextra_is_clawfree_vpar");
#endif

  memset (degrees+1, '\0', n*sizeof(degrees[0]));
  int v;
  for (v = 1; v <= n; v++) {
    int parent = vpar[v];
    assert (0 <= parent && parent <= n);
    if (parent) {
      if (++degrees[v] >= 3 || ++degrees[parent] >= 3)
        return FALSE;
    }
  }
  return TRUE;
}

/* Search for induced claws in g,m,n.
   g must be undirected (edges both ways).  Any self-loops are ignored.
   Each claw found is

       b
        \          induced subgraph, for some vertices v,a,b,c
         v---a     with a < b < c,
        /          no edges a-b, b-c or c-a
       c

   and calls with the 4 vertices and user "data" pointer,

       stop = (*callback) (v,a,b,c, data);

   If (*callback) returns 0 then the search continues.
   If (*callback) returns non-0 then the search stops and
   nautyextra_claw_search() returns with that non-0 value.
   If the search runs to completion then nautyextra_claw_search() returns 0.

   The arms are ordered a < b < c in each callback.  v can be any vertex
   number, but is different from a,b,c of course.

   The current code is a naive search over all vertices v, looking for
   suitable neighbours a,b,c.  This means 3 neighbours with no edges
   between.  v can have more then 3 neighbours.  All suitable sets of 3 of
   them can be induced claws.
*/
int
nautyextra_claw_search (const graph *g, int m, int n,
                        int (*callback)(const graph*,int,int,
                                        int,int,int,int, void*),
                        void *data)
{
  /* ENHANCE-ME: Find b by loop over rows vword & ~aword, so bit masks of
     those with v-b and not a-b.  Likewise c by v & ~a & ~b.
  */
  DEBUG(printf("nautyextra_claw_search() %p n=%d\n", g, n));
  assert (n >= 0);

  set *v_row = (set*) g;
  int v;
  for (v = 0; v < n; v++, v_row += m) {
    DEBUG(printf(" v=%d\n", v));
    int a;
    for (a = -1; (a = nextelement(v_row,m,a)) >= 0; ) {
      if (a == v) continue;  /* ignore self-loop */
      set *a_row = GRAPHROW(g,a,m);

      int b;
      for (b = a; (b = nextelement(v_row,m,b)) >= 0; ) {
        if (b == v) continue;  /* ignore self-loop */
        if (ISELEMENT(a_row,b)) continue;  /* a--b edge not claw */
        set *b_row = GRAPHROW(g,b,m);

        int c;
        for (c = b; (c = nextelement(v_row,m,c)) >= 0; ) {
          if (c == v) continue;  /* ignore self-loop */
          if (ISELEMENT(a_row,c)) continue;  /* a--c edge not claw */
          if (ISELEMENT(b_row,c)) continue;  /* b--c edge not claw */

          assert (a < b && b < c && a!=v && b!=v && c!=v);
          int ret = (*callback)(g,m,n, v,a,b,c, data);
          if (ret) return ret;
        }
      }
    }
  }
  return 0;
}

/* nautyextra_claw_count() return the number of induced claws in g,m,n.
   g must be undirected (arcs both directions).  Any self-loops are ignored.
*/
static int
INTERNAL_claw_count_callback (const graph *g, int m, int n,
                     int v, int a, int b, int c, void *data)
{
  (*(NAUTYEXTRA_TYPE_ULL*)data)++;
  DEBUG(printf("  claw %d to %d,%d,%d  (count %d)\n", v,a,b,c, *(int*)data));
  return 0;
}
NAUTYEXTRA_TYPE_ULL
nautyextra_claw_count (const graph *g, int m, int n)
{
  NAUTYEXTRA_TYPE_ULL count = 0;
  nautyextra_claw_search (g,m,n, &INTERNAL_claw_count_callback, &count);
  return count;
}

/* nautyextra_is_clawfree() returns TRUE if g is claw-free, meaning does not
   contain a claw as an induced sugraph, or returns FALSE if does have a claw.
*/
static int
INTERNAL_is_clawfree_callback (const graph *g, int m, int n,
                                 int v, int a, int b, int c, void *data)
{
  /* any claw, stop search */
  return 1;
}
int
nautyextra_is_clawfree (const graph *g, int m, int n)
{
  /* nautyextra_claw_search() return 0 on search ending, or
     nautyextra_is_clawfree_callback() return 1 on find a claw */
  return ! nautyextra_claw_search (g,m,n,
                                   &INTERNAL_is_clawfree_callback, NULL);
}


/* ------------------------------------------------------------------------- */
/* Degrees */

/* Set degrees[0..n-1] so degrees[v] = degree of vertex v in g,m,n.
   g must be undirected (arcs both ways).
   A self-loop counts as 1 for the degree.
   degrees[] must have space for n many ints.
 */
void
nautyextra_degrees_array (const graph *g, int m, int n, int *degrees)
{
  assert (n >= 0);
  memset (degrees, '\0', n*sizeof(degrees[0]));
  while (n--) {
    *degrees++ = setsize((set*)g, m);
    g += m;
  }
}

/* Set degrees[1..n] so degrees[v] = the degree of vertex v in tree vpar[1..n].
   degrees[] must have space for n+1 many ints.
   degrees[0] is set to 0 (there being no such vertex in a vpar).
 */
void
nautyextra_degrees_array_vpar (const int *vpar, int n, int *degrees)
{
  int v;
  assert (n >= 0);
  memset (degrees, '\0', (n+1)*sizeof(degrees[0]));
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    assert (0 <= p && p <= n);
    if (p) {
      degrees[v]++;
      degrees[p]++;
    }
  }
}

/* ------------------------------------------------------------------------- */
/* Vpar Paths and Ancestry */

/* Return the root vertex of v in tree vpar[1..n].
   This goes up parent vpar[v], vpar[vpar[v]], etc, until reaching a vertex
   with vpar[r]==0, that being the root r.
   There is no n parameter as the search just continues up until vpar[r]==0.
   vpar can be a forest, in which case the return is the root is of the
   component tree containing v.
*/
int
nautyextra_root_of_vertex_vpar (const int *vpar, int v)
{
  assert (v >= 1);
  while (vpar[v]) v = vpar[v];
  return v;
}

/* Return the number of roots in vpar[1..n].
   This is simply how many vpar[v]==0.
   For a tree this is 1 (when n>0) but for a forest is more.
*/
int
nautyextra_num_roots_vpar (const int *vpar, int n)
{
  assert (n >= 0);
  int count = 0;
  int v;
  for (v = 1; v <= n; v++)
    count += (vpar[v]==0);
  return count;
}

/* g is a directed graph.
   Return the number of arcs in g, including self-loop arcs.
   Cf degstats() which gives edges of an undirected graph, and more.
*/
int
nautyextra_num_arcs (const graph *g, int m, int n)
{
  return setsize((set*)g, n*m);
}

/* Return the number of edges in tree vpar[1..n].
   This is the number of vpar[v]!=0.
   For a tree this is n-1 (or 0 if n==0), but for a forest is less.

   Cf for graph degstats(), giving edges and more.
*/
int
nautyextra_num_edges_vpar (const int *vpar, int n)
{
  return n - nautyextra_num_roots_vpar(vpar,n);
}

/* Return TRUE if "ancestor" is an ancestor of v in vpar[1..n], or FALSE if not.
   If v==ancestor then return TRUE, so the test is whether v is in the
   subtree starting at ancestor.
 */
int
nautyextra_is_ancestor_vpar (const int *vpar, int v, int ancestor)
{
  DEBUG(printf("nautyextra_is_ancestor_vpar() %d to %d\n", v, ancestor));
  assert (ancestor >= 1);
#ifndef NDEBUG
  int count = 0;
  int n = 0;
#endif
  do {
    assert (v >= 1);
    assert ((n = MAX(n,v+3)));
    assert (++count < n);    /* guard against infinite loop */

    if (v == ancestor) {
      DEBUG(printf("  yes\n"));
      return TRUE;
    }
    v = vpar[v];

  } while (v);
  DEBUG(printf("  no\n"));
  return FALSE;
}

/* Return the depth (distance from its root) of vertex v in tree vpar[1..n].
 */
int
nautyextra_depth_of_vertex_vpar (const int *vpar, int v)
{
  DEBUG(printf("nautyextra_depth_of_vertex_vpar() %d\n", v));
  int depth = -1;
  while (v) {
    assert (v >= 1);
    v = vpar[v];
    depth++;
  }
  DEBUG(printf(" return depth %d\n", depth));
  return depth;
}

/* Return the common ancestor vertex of u and v in vpar[1..n].
   This might be u or v itself if one is ancestor of the other.
   If vpar is a forest and u,v are in different component trees then return 0.
*/
int
nautyextra_common_ancestor_vpar (const int *vpar, int u, int v)
{
  /* ENHANCE-ME: Could record vertices seen working up from u,v and stop on
     revisit rather than go all the way to the root.  But perhaps memset of
     flags or a depths array would be slower than just running up.
  */

  int u_depth = nautyextra_depth_of_vertex_vpar(vpar,u);
  int v_depth = nautyextra_depth_of_vertex_vpar(vpar,v);
  while (u_depth > v_depth) {
    assert (vpar[u] != 0);
    u = vpar[u];
    u_depth--;
  }
  while (v_depth > u_depth) {
    assert (vpar[v] != 0);
    v = vpar[v];
    v_depth--;
  }
  while (u != v) {
    u = vpar[u];
    v = vpar[v];
    if (u==0) {
      assert (v==0);
      return 0;    /* no common ancestor */
    }
    assert(v != 0);
  }
  return u;
}

/* Return the path length from u to v in vpar[1..n].
   (There is no n parameter, since u and v are simply follows upwards to a
   common ancestor.)
   If u==v then return 0, for 0 distance between.
   If u and v are in different component trees of a forest then return -1.
*/
int
nautyextra_path_length_vpar (const int *vpar, int u, int v)
{
  DEBUG(printf("nautyextra_path_length_vpar() %d -- %d\n", u,v));

  /* ENHANCE-ME: For a short path in a deep tree it might be better to track
     the tree vertices visited to notice u or v common ancestor.  But maybe
     big memset of flags or depths would be slower usually.  Sometimes some
     2x, 1x stepping would have the 2x hit the 1x of the other, but not
     always.
  */

  int path_length = 0;
  int u_depth = nautyextra_depth_of_vertex_vpar(vpar,u);
  int v_depth = nautyextra_depth_of_vertex_vpar(vpar,v);
  while (u_depth > v_depth) {
    assert (vpar[u] != 0);
    u = vpar[u];
    path_length++;
    u_depth--;
  }
  while (v_depth > u_depth) {
    DEBUG(printf(" v_depth=%d u_depth=%d up\n", u_depth, v_depth));
    assert (vpar[v] != 0);
    v = vpar[v];
    path_length++;
    v_depth--;
  }
  while (u != v) {
    u = vpar[u];
    v = vpar[v];
    if (u==0) {
      assert (v==0);
      return -1;    /* no common ancestor */
    }
    assert(v != 0);
    path_length += 2;
  }
  return path_length;
}

/* Set path[] to the vertices of path u to v in vpar[1..n].
   (There is no n parameter, since u and v are simply follows upwards to a
   common ancestor.)
   The return value is path length "len", same as nautyextra_path_length_vpar().
   path[] must have space for vertices path[0..len].
   The longest path would be through all n vertices, for len=n-1 and path[]
   needing space for n ints.
*/
int
nautyextra_path_vertices_vpar (const int *vpar, int u, int v, int *path)
{
  DEBUG(printf("nautyextra_path_vertices_vpar() %d->%d\n", u,v));

  int num = 0;
  int v_length = 0;
  int u_depth = nautyextra_depth_of_vertex_vpar(vpar,u);
  int v_depth = nautyextra_depth_of_vertex_vpar(vpar,v);
  while (u_depth > v_depth) {
    assert (vpar[u] != 0);
    DEBUG(printf(" u deeper store %d\n", u));
    *path++ = u; num++;
    u = vpar[u];
    u_depth--;
  }
  int orig_v = v;
  while (v_depth > u_depth) {
    DEBUG(printf(" v_depth=%d u_depth=%d up\n", u_depth, v_depth));
    assert (vpar[v] != 0);
    v = vpar[v];
    v_length++;
    v_depth--;
  }
  while (u != v) {
    DEBUG(printf(" u side store %d\n", u));
    *path++ = u; num++;
    u = vpar[u];
    v = vpar[v];
    if (u==0) {
      assert (v==0);
      return 0;    /* no common ancestor */
    }
    assert(v != 0);
    v_length++;
  }
  *path++ = u; num++;   /* common ancestor */
  DEBUG(printf(" common ancestor store %d, num now %d\n", u, num));
  path += v_length;  num += v_length;

  v = orig_v;
  while (v != u) {
    *--path = v;
    v = vpar[v];
  }
  num--;  /* to path length */
  assert (num >= 0);
  return num;
}


/* ------------------------------------------------------------------------- */
/* Eccentricity

   Cf distvals() for sparsegraph all distances from v in a sparsegraph (the
   max of which is eccentricity).  distances() or distances_sg() is
   histogram of how many at distance, with some cell stuff.
*/


/* Return the eccentricity of vertex target_v in tree vpar[1..n].
   This is the distance to the vertex furthest away from target_v.
   vpar can be a forest, in which case the eccentricity is to the furthest
   vertex within the component tree of target_v.

   The approach is to work upwards from leaves, forming at each v the
   eccentricity of v into the subtree below it.  This is max(c_ecc+1) for
   each child of v.  When target_v is reached the subtree eccentricity is an
   initial candidate for the return value.

   For an ancestor vertex p above target_v, the distance above_target is
   propagated up.  New candidate eccentricities are distance above_target
   plus the subtree eccentricity below p, not including the subtree
   containing the target itself.

   The leaf-order traversal means the children of an ancestor might be
   accumulated into p in any order.  Before p is known to be an ancestor
   they establish an ecc_below at p.  When the target_v subtree is found at
   p it considers that ecc_below so far plus above_distance.  Subsequent
   subtrees at p consider the new child ecc_below plus the distance_above.
*/
int
nautyextra_eccentricity_of_vertex_vpar (const int *vpar, int n, int target_v)
{
  DEBUG(printf("nautyextra_eccentricity_of_vertex_vpar() n=%d v=%d\n",
               n, target_v);
        nautyextra_print_vpar(vpar,n));
  assert(n >= 0);
  assert(1 <= target_v && target_v <= n);

  int i;
  int eccentricity = 0;
  struct data {
    int num_children;
    int ecc_below;
    int above_target;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_eccentricity_of_vertex_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      int c_ecc_below    = data[v].ecc_below;
      int c_above_target = data[v].above_target;
      int c_is_target    = (v == target_v);
      assert (data[v].num_children == 0);

      DEBUG(printf("consider v=%d ecc_below %d above_target %d\n",
                   v, c_ecc_below, c_above_target));

      if (c_is_target)
        eccentricity = c_ecc_below;

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      assert(data[v].num_children > 0);
      if (! v) break;  /* root */

      int new_ecc        = c_ecc_below + 1;
      int v_ecc_below    = data[v].ecc_below;  /* before this child */
      int v_above_target = data[v].above_target;
      DEBUG(printf(" parent %d ecc_below %d above_target %d\n",
                   v, data[v].ecc_below, data[v].above_target));
      assert (! (v_above_target && (c_is_target || c_above_target)));

      if (v_above_target) {
        int target_ecc = v_above_target + new_ecc;
        DEBUG(printf(" parent %d above target, possible ecc %d+%d=%d%s\n",
                     v, v_above_target, new_ecc, target_ecc,
                     target_ecc > eccentricity?" new high":" not new high"));
        eccentricity = MAX(eccentricity, target_ecc);

      } else if (c_is_target || c_above_target) {
        data[v].above_target = v_above_target = c_above_target + 1;
        int target_ecc = v_ecc_below + v_above_target;
        eccentricity = MAX(eccentricity, target_ecc);
        DEBUG(printf(" c target or above, onto parent %d ecc %d+%d+1 = %d\n",
                     v, v_ecc_below,v_above_target, target_ecc));
      }

      if (new_ecc > v_ecc_below) {
        DEBUG(printf(" parent %d new high ecc %d\n", v, new_ecc));
        data[v].ecc_below = new_ecc;
      } else {
        DEBUG(printf(" parent %d not new high ecc\n", v));
      }

      if (--data[v].num_children) break;
    }
  }
  return eccentricity;
}


/* ------------------------------------------------------------------------- */
/* Diameter */

/* Return the diameter of tree vpar[1..n].
   vpar should be a single tree, not a forest.
 */
int
nautyextra_diameter_vpar (const int *vpar, int n)
{
  struct nautyextra_diameter_info info;
  nautyextra_diameter_info_vpar (vpar, n, &info);
  return info.diameter;
}

/* Calculate struct nautyextra_diameter_info for tree vpar[1..n].

   The approach is to work up from the leaf vertices.  The fields calculated
   are

       ecc_below          eccentricity of subtree below v
       ecc_v              some vertex attaining the eccentricity
       ecc_endpoints      number of vertices of depth ecc_below
       ecc_vertices       number of vertices on paths to ecc_below
       above_diameter     diameter value which v is above
       above_distance     distance above that diameter

   Eccentricity at a vertex v follows from its children

       ecc_below = 1 + max(c ecc_below)
       ecc_v     = any of c ecc_v with maximal ecc_below,
                   or v itself if no children
       ecc_endpoints = sum c ecc_endpoints, or 1 for v itself if no children

   At a new high max(c ecc_below) the ecc_endpoints and ecc_vertices reset
   to that child, otherwise they accumulate.

   A diameter candidate is eccentricities between child sub-tree and parent
   sub-tree, before that child has been considered in the parent.

       new_diameter = c ecc_below + v ecc_below + 1

   If this is a new high then new return values are established

       diameter v1,v2     = c ecc_v, v ecc_v
       diameter p         = v, the common ancestor
       diameter endpoints = c ecc_endpoints + v ecc_endpoints
       diameter vertices  = c ecc_vertices  + v ecc_vertices

   If this is an equal high there are two cases.  If v is already part of
   the diameter, or somewhere above the diameter, then accumulate the child

       diameter endpoints += c ecc_endpoints
       diameter vertices  += c ecc_vertices + v above_diameter - 1

   If c is already part of the diameter or somewhere above then conversely
   accumulate the parent

       diameter endpoints += v ecc_endpoints
       diameter vertices  += v ecc_vertices + c above_diameter - 1

   above_diameter is used to determine which of these is the case.
   data[v].above_diameter is the diameter of the subtree at and below v.  If
   this is equal to the whole tree diameter (so far) then the diameter
   endpoints and vertices found in there become part of the total.

   It's possible neither c nor v is above diameter of the whole so far.
   This occurs when some path length < diameter is encountered in different
   subtrees.  Can do either or neither of the accumulation cases, since
   there will be a new high diameter replacing these values later anyway.

   Both new high diameter and equal high diameter set v to the diameter at
   that point.

       v above_distance = 0
       v above_diameter = diameter

   Otherwise, for prospective < diameter at v, distance above the diameter
   propagates up,

       if c above_diameter == whole tree diameter
           v above_diameter = c above_diameter
           v above_distance = c above_distance + 1

   An alternative for this is just to propagate when c above_diameter >=
   v_above_diameter, but it's convenient to compare to the whole tree
   diameter and let v above_diameter be only written in the accumulation,
   not read and compared.

   Total work is the usual 2 to 3 passes over vpar.
*/
void
nautyextra_diameter_info_vpar (const int *vpar, int n, struct nautyextra_diameter_info *info)
{
  DEBUG(printf("nautyextra_diameter_info_vpar() n=%d\n", n));
  assert(n >= 0);

  int i;
  int diameter = (n == 0 ? -1 : 0);
  int diameter_v1 = n;
  int diameter_v2 = n;
  int diameter_p  = n;
  int diameter_endpoints = n;
  int diameter_vertices  = n;
  struct data {
    int num_children;
    int ecc_below;
    int ecc_v;                 /* a vertex attaining ecc_below */
    int ecc_endpoints;
    int ecc_vertices;
    int above_diameter;
    int above_distance;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_diameter_info_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].ecc_v              = i;   /* v itself */
    data[i].ecc_endpoints      = 1;
    data[i].ecc_vertices       = 1;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children != 0) continue;

    for (;;) {
      int new_ecc              = data[v].ecc_below + 1;
      int c_ecc_v              = data[v].ecc_v;
      int c_ecc_endpoints      = data[v].ecc_endpoints;
      int c_ecc_vertices       = data[v].ecc_vertices;
      int c_above_diameter     = data[v].above_diameter;
      int c_above_distance     = data[v].above_distance;
      DEBUG(printf("consider v=%d ecc_below %d endpoints %d vertices %d, above_diameter %d,%d\n",
                   v, data[v].ecc_below, c_ecc_endpoints, c_ecc_vertices,
                   c_above_diameter, c_above_distance));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      assert(data[v].num_children > 0);
      if (! v) break;

      int new_diameter     = data[v].ecc_below + new_ecc;
      int v_ecc_endpoints  = data[v].ecc_endpoints;
      int v_ecc_vertices   = data[v].ecc_vertices;
      int v_above_distance = data[v].above_distance;
      DEBUG(printf("  new_diameter %d cf %d\n", new_diameter, diameter));

      if (new_diameter >= diameter) {
        if (new_diameter > diameter) {
          diameter    = new_diameter;
          diameter_v1 = c_ecc_v;
          diameter_v2 = data[v].ecc_v;
          diameter_p  = v;
          diameter_endpoints = c_ecc_endpoints + v_ecc_endpoints;
          diameter_vertices  = c_ecc_vertices  + v_ecc_vertices;
          DEBUG(printf("  new high diameter v=%d,%d to p=%d endpoints %d, vertices %d\n",
                       diameter_v1, diameter_v2, diameter_p,
                       diameter_endpoints, diameter_vertices));
        } else {
          int diameter_new_endpoints
            = (c_above_diameter==diameter ? v_ecc_endpoints : c_ecc_endpoints);
          diameter_endpoints += diameter_new_endpoints;
          int diameter_new_vertices
            = (c_above_diameter==diameter
               ? v_ecc_vertices + c_above_distance
               : c_ecc_vertices + v_above_distance);
          diameter_vertices  += diameter_new_vertices;
          DEBUG(printf("  equal high diameter, add endpoints %d to %d, vertices add %d to %d\n",
                       diameter_new_endpoints, diameter_endpoints,
                       diameter_new_vertices,  diameter_vertices));
        }
        data[v].above_distance = 0;
        data[v].above_diameter = diameter;

      } else {
        if (c_above_diameter==diameter) {
          data[v].above_diameter = c_above_diameter;
          data[v].above_distance = c_above_distance + 1;
          DEBUG(printf("  parent above diameter +1 to %d,%d\n",
                       data[v].above_diameter, data[v].above_distance));
        }
      }

      int v_ecc_below = data[v].ecc_below;
      if (new_ecc > v_ecc_below) {
        data[v].ecc_below = new_ecc;
        data[v].ecc_v = c_ecc_v;
        data[v].ecc_endpoints = c_ecc_endpoints;
        data[v].ecc_vertices  = c_ecc_vertices + 1;
        DEBUG(printf("  new high parent ecc %d endpoints %d\n",
                     data[v].ecc_below, data[v].ecc_endpoints));
      } else if (new_ecc == v_ecc_below) {
        data[v].ecc_endpoints += c_ecc_endpoints;
        data[v].ecc_vertices  += c_ecc_vertices;
      }

      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d  (diameter %d)\n", v, diameter));
        break;
      }
    }
  }

  assert (diameter_endpoints >= (n<2 ? n : 2));
  assert (diameter_endpoints <= diameter_vertices);
  assert (diameter_vertices >= (n<2 ? n : 2));
  assert (diameter_vertices <= n);

  /* C---C          diameter=1, radius=1
     .---C---.      diameter=2, radius=1
     .---C---C---.  diameter=3, radius=2
  */
  DEBUG(printf("return v=%d,%d to p=%d diameter %d endpoints %d vertices %d\n",
               diameter_v1, diameter_v2, diameter_p, diameter,
               diameter_endpoints, diameter_vertices));
  info->diameter      = diameter;
  info->radius        = (diameter+1) >> 1;
  info->end_v1        = diameter_v1;
  info->end_v2        = diameter_v2;
  info->end_ancestor  = diameter_p;
  info->num_endpoints = diameter_endpoints;
  info->num_vertices  = diameter_vertices;
}


/* ------------------------------------------------------------------------- */
/* Subtree Sizes and Centroid */

/* Return the size (number of vertices) of the subtree at and below v in
   tree vpar[1..n].  The size includes v itself.
 */
int
nautyextra_subtree_size_vpar (const int *vpar, int n, int v)
{
  int i;
  struct data {
    int num_children;
    int num_below;     /* number of vertices below, not including self */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_subtree_size_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_subtree_size_vpar()\n"));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
  }

  for (i = 1; i <= n; i++) {
    DEBUG(printf("i=%d\n", i));
    int l = i;
    if (data[l].num_children) {
      /* either num_children=-1 for l already considered,
         or num_children>0 for l not yet a leaf */
      continue;
    }
    for (;;) {
      DEBUG(printf("consider l=%d num_below %d\n", l, data[v].num_below));
      int b = data[l].num_below + 1;
      assert (data[l].num_children == 0);

      if (l == v) {
        DEBUG(printf("  is target, return %d\n", b));
        return b;
      }
      data[l].num_children = -1;
      l = vpar[l];     /* parent */
      if (! l) break;
      data[l].num_below += b;

      assert(data[l].num_children > 0);
      if (--data[l].num_children) break;
    }
  }
  return 0;
}

/* Set num_children[1..n] so num_children[v] is the number of children of v.
 */
void
nautyextra_num_children_array_vpar (const int *vpar, int n, int *num_children)
{
  int v;
  assert (n >= 0);
  assert (num_children != NULL);

  memset(num_children, '\0', (n+1)*sizeof(num_children[0]));
  for (v = 1; v <= n; v++)
    num_children[vpar[v]]++;
}

/* Set sizes[1..n] so sizes[v] is the size (number of vertices) of the
   subtree at and below v in vpar[1..n].  Each size includes v itself.

   These sizes are the same as nautyextra_subtree_size_vpar(), but
   calculated for the whole tree at once.
 */
void
nautyextra_subtree_sizes_array_vpar (const int *vpar, int n, int *sizes)
{
  assert (n >= 0);
  assert (sizes != NULL);
#if MAXN
  int num_children[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(int,num_children,n+1,"nautyextra_subtree_sizes_array_vpar");
#endif

  nautyextra_num_children_array_vpar(vpar,n,num_children);
  int i;
  for (i = 1; i <= n; i++)
    sizes[i] = 1;

  for (i = 1; i <= n; i++) {
    int v = i;
    if (num_children[v]) continue;
    for (;;) {
      int size = sizes[v];
      assert (num_children[v] == 0);

      num_children[v] = -1;
      v = vpar[v];     /* parent */
      if (! v) break;
      sizes[v] += size;

      assert(num_children[v] > 0);
      if (--num_children[v]) break;
    }
  }
}

/* Set *info to information about the centroid of tree vpar[1..n].
 */
void
nautyextra_centroid_info_vpar (const int *vpar, int n,
                               struct nautyextra_centroid_info *info)
{
  DEBUG(printf("nautyextra_centroid_info_vpar() %d\n", n);
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int num_children;
    int num_below;  /* number of vertices below v, not including v itself */
    int weight;     /* max size of subtrees */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_centroid_info_vpar");
#endif
  int centroid_v1 = 0;
  int centroid_v2 = 0;
  int centroid_weight = n;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  for (i = 1; i <= n; i++) {
    DEBUG(printf("i=%d\n", i));
    int v = i;
    if (data[v].num_children) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }
    for (;;) {
      int new_num_below = data[v].num_below + 1;   /* including v */
      int p_subtree_size = n - new_num_below;
      DEBUG(printf("consider v=%d parent %d  weight %d num_below %d b=%d  p_subtree_size %d\n",
                   v, vpar[v], data[v].weight, data[v].num_below,
                   new_num_below, p_subtree_size));
      assert (data[v].num_children == 0);

      int weight = data[v].weight;
      weight = MAX(weight, p_subtree_size);
      DEBUG(printf(" weight %d\n", weight));

      if (weight < centroid_weight) {
        DEBUG(printf(" new low weight for centroid\n"));
        centroid_v1 = v;
        centroid_v2 = v;
        centroid_weight = weight;
      } else if (weight == centroid_weight) {
        DEBUG(printf(" equal low weight for centroid\n"));
        assert (v != centroid_v1);
        if (v < centroid_v1)
          centroid_v1 = v;
        else
          centroid_v2 = v;
      }

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) break;

      data[v].num_below += new_num_below;
      int p_weight = data[v].weight;
      data[v].weight = MAX(p_weight,new_num_below);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        break;
      }
    }
  }
  info->v1 = centroid_v1;
  info->v2 = centroid_v2;
  info->weight = centroid_weight;
}



/* ------------------------------------------------------------------------- */
/* Independent Sets */

/* An independent set ("indset") is a set of vertices which do not have any
   edge between them, ie. none are immediate neighbours.

   The number of independent sets is often large, being for example 2^n for
   totally disconnected n vertices.

   Cf nauty indsets(), cellind() something with cells.
*/

/* g,m,n is an undirected graph.
   ss is a set of n vertices.
   Return TRUE if ss is an independent set in g.  Return FALSE if not.
   This means checking each v in ss has none of its neighbours in ss.
*/
int
nautyextra_is_indset (const graph *g, int m, int n, const set *ss)
{
  if (n == 0) return TRUE;  /* n=0 empty set ss is independent */
  int v;
  for (v = -1; (v = nextelement((set*)ss,m,v)) >= 0; ) {
    const setword *sp = ss;
    setword *row = GRAPHROW(g,v,m);
    int i;
    for (i = 0; i < m; i++)
      if (*row++ & *sp++)
        return FALSE;  /* v in ss, and also some neighbour of v in ss */
  }
  return TRUE;

  /* Equivalently in the loop, but here can stop at any intersection, rather
     than full count.  if (setinter(row, (set*)ss, m)) return FALSE;
  */
}

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1 inclusive.
   Return TRUE if ss is an independent set in sg.  Return FALSE if not.
   This means checking each v in ss has none of its neighbours in ss.

   sg can have self-loops.  If ss includes any vertex with a self-loop then
   it is not independent.  sg can have multi-edges.
*/
int
nautyextra_is_indset_sg (const sparsegraph *sg, const set *ss)
{
  int n = sg->nv;
  int m = SETWORDSNEEDED(n);
  if (m == 0) return TRUE;  /* n=0 empty set ss is independent */
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int from;
  for (from = -1; (from = nextelement((set*)ss,m,from)) >= 0; ) {
    const int *eptr = ee + vv[from];
    int degree = dd[from];
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      if (ISELEMENT(ss,to)) return FALSE;  /* from,to both in ss */
    }
  }
  return TRUE;
}

/* Return the number of independent sets in sg.

   Currently this is a naive search of all sets, so is suitable only for few
   vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE many vertices.

   ENHANCE-ME: Prune like nautyextra_indnum_and_count() below.
   ENHANCE-ME: Taking vertices in a neighbour-first order would then help
   prune more too.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_indsets_count_sg (const sparsegraph *sg)
{
  setword count = 0;
  int n = sg->nv;

  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_indsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  setword ss;
  NAUTYEXTRA_SET1_LOOP(ss,n, count += nautyextra_is_indset_sg(sg,&ss));
  return count;
}

/* Return the number of independent sets in tree vpar[1..n].
   vpar can be a forest.  In that case the count is product of counts in
   each component tree.
   The count can be large, up to about 2^(n-1).
   Currently overflow results in an abort().

   At a given vertex v there is a count of how many sets there are for the
   sub-tree at that vertex and below, separated into counts of how many
   which include v and how many do not.

   The sets which include vertex v cannot include any of its children.  The
   sets which do not include v can either include or not include the
   children, so

       with    = prod           c_without
       without = prod (c+with + c_without)

   These products are built going upwards in the tree from the leaf vertices.
   Initially with=1 and without=1, being a single set with or without v.
 */
NAUTYEXTRA_TYPE_ULL
nautyextra_indsets_count_vpar (const int *vpar, int n)
{
  if (n == 0) return 1;
  if (n == 1) return 2;

  /* 1--2        sets empty 1 2       = 3
     3-->1<--2   sets empty 1 2 3 32  = 5
  */

  int i;
  struct data {
    int                  num_children;
    NAUTYEXTRA_TYPE_ULL  with;
    NAUTYEXTRA_TYPE_ULL  without;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indsets_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indsets_count_vpar()");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with    = 1;
    data[i].without = 1;
  }

  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d count %Lu,%Lu\n",
                 v, data[v].num_children,
                 (unsigned long long) data[v].with,
                 (unsigned long long) data[v].without));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      NAUTYEXTRA_TYPE_ULL c_with    = data[v].with;
      NAUTYEXTRA_TYPE_ULL c_without = data[v].without;
      NAUTYEXTRA_TYPE_ULL c_with_and_without;
      ADD_NO_OVERFLOW_ULL(c_with_and_without, c_without, c_with);

      DEBUG(printf("consider v=%d count %Lu,%Lu = %Lu\n",
                   v, c_with, c_without, c_with_and_without));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  root\n"));
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_with_and_without);
        break;
      }

      MUL_NO_OVERFLOW_ULL(data[v].with,
                          data[v].with, c_without);

      MUL_NO_OVERFLOW_ULL(data[v].without,
                          data[v].without, c_with_and_without);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  return total_count;
}


/* ------------------------------------------------------------------------- */
/* Independence Number */

/* The independence number ("indnum") of a graph is the number of vertices
   in its largest independent set.

   A set of this size is a maximal independent set, but it's possible for
   some maximal sets to be smaller.
*/

/* g is a directed or undirected graph.
   Set *info to its independence number and count of how many sets have that
   size.
   This is currently an exhaustive search so suitable only for small n.

   The search is pruned when a conflicting new vertex is found, and the
   search is further restricted to sets big enough to at least equal the
   current indnum so far.  This can help speed-up the search, especially
   when the graph has many edges.

   ENHANCE-ME: Searching in a neighbours-first order would exclude
   conflicting vertices sooner.
   ENHANCE-ME: Any leaf vertices should be always included, since between a
   leaf and its attachment, a set of >= size is obtained by taking the leaf.
*/
void
nautyextra_indnum_and_count (const graph *g, int m, int n,
                             struct nautyextra_indnum_and_count *info)
{
  DEBUG(printf("nautyextra_indnum_and_count() g=%p m=%d,n=%d to %p\n",
               g,m,n,info));

  if (n <= 1) {
    info->indnum = n;
    info->count = 1;
    DEBUG(printf("  return %d, %Lu\n",
                 info->indnum, (unsigned long long) info->count));
    return;
  }
#if MAXN
  set ss[SETWORDSNEEDED(MAXN)];
  assert(n <= MAXN);
#else
  WORKSPACE(setword,ss,m,"nautyextra_indnum_and_count");
#endif
  assert (n >= 0);
  assert (m >= 0);

  int                  indnum = 0;
  NAUTYEXTRA_TYPE_ULL  count = 1;
  if (n == 0) goto END;

  int v = 0;
  int ss_size = 0;
  EMPTYSET(ss,m);
  for (;;) {
    DEBUG(printf("try add v=%d to ss=%s (size %d)\n",
                 v, nautyextra_set_str(ss,n), ss_size));
    assert (v >= 0);
    assert (v < n);
#ifndef NDEBUG
    { int i; for (i=0; i<=v; i++) assert (! ISELEMENT(ss,i)); }
#endif
    assert (ss_size == setsize(ss,m));
    assert (indnum >= ss_size);
    assert (ss_size+v+1 >= indnum);

    ADDELEMENT(ss,v);
    ss_size++;
    assert (ss_size == setsize(ss,m));

    set *gv = GRAPHROW(g,v,m);
    int i;
    for (i=0; i<m; i++) {
      if (gv[i] & ss[i]) {
        DEBUG(printf("  have edge v row %s (%p)\n", nautyextra_set_str(gv,n), gv);
              printf("               ss %s\n", nautyextra_set_str(ss,n)));
        goto BACKTRACK;
      }
    }
    DEBUG(printf("  no edge v row %s (%p)\n", nautyextra_set_str(gv,n), gv);
          printf("             ss %s\n", nautyextra_set_str(ss,n)));
    if (ss_size > indnum) {
      DEBUG(printf("  new biggest indnum %d\n", ss_size));
      indnum = ss_size;
      count = 0;

      /* if (NAUTYEXTRA_PROGRESS) */
      /*   printf("indnum %d\r", indnum); */
    }
    count += (ss_size == indnum);
    DEBUG(if (ss_size == indnum) printf("  count now %Lu\n", count));

    /* if (ss_size == 12) { */
    /*   printf("ss %s\n", nautyextra_set_str(ss,n)); */
    /* } */

    if (v) {
      /* Need a further indnum-ss_size vertices in ss for it to be >= the
         indnum found so far.  Going to v=indnum-ss_size-1 will be exactly
         indnum-ss_size many to find another equal to indnum.  If have
         indnum==ss_size because the current set was in fact another equal
         then this is v=-1, and in that case go to v=0 (which if independent
         will give a new higher indnum).  Since current v>0, this 0 vertex
         available to try.
      */
      v = indnum - ss_size - 1;
      DEBUG(printf("  jump down to v=%d  (indnum=%d ss_size=%d)\n",
                   v, indnum, ss_size));
      assert(v >= -1);
      if (v < 0) v = 0;
      continue;
    }
    DEBUG(printf("  low end\n"));

  BACKTRACK:
    DEBUG(printf("backtrack seek next highest 0\n"));
    for (;;) {
      assert (ISELEMENT(ss,v));
      DELELEMENT(ss,v);
      ss_size--;
      assert (ss_size == setsize(ss,m));

      if (++v >= n) {
        DEBUG(printf("end\n"));
        goto END;
      }
      DEBUG(printf("backtrack to v=%d in ss=%s\n",
                   v, nautyextra_set_str(ss,m)));
      if (! ISELEMENT(ss,v)) {
        DEBUG(printf("  not in ss\n"));
        break;
      }
    }
  }

 END:
  info->indnum = indnum;
  info->count = count;
}

/* Return the independence number of tree vpar[1..n].
   vpar can be a forest, in which case the effect is sum indnums of its
   component trees.

   The code takes a leaf into the independent set, deletes it, and reckons
   its parent as excluded.  Then repeat, taking any leaf not excluded and
   excluding its parent.  A parent with multiple children might be excluded
   more than once.  Any exclusion means omit it.

   Between a leaf and its parent, taking the leaf and excluding the parent
   gives a set size >= taking the parent.
 */
int
nautyextra_indnum_vpar (const int *vpar, int n)
{
  int indnum = 0;
  int i;
  struct data {
    int  num_children;
    int  exclude;        /* 0 or 1 */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indnum_vpar() n=%d", n);
        nautyextra_print_vpar(vpar,n));

  if (n==0) return 0;
  if (n==1) return 1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n",
                 v, data[v].num_children));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_exclude = data[v].exclude;
      DEBUG(printf("consider v=%d (parent %d) exclude %d  (indnum %d)\n",
                   v, vpar[v], c_exclude, indnum));
      assert (data[v].num_children == 0);

      /* take leaf whenever not excluded */
      int take_c = 1 - c_exclude;
      indnum += take_c;

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) break;  /* root */

      /* exclude parent when child is taken */
      data[v].exclude |= take_c;

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  DEBUG(printf("  final %d\n", indnum));
  return indnum;
}

/* Set *info to the independence number and count of sets of this size in
   tree vpar vpar[1..n].
   vpar can be a forest, in which case the effect is sum indnums and product
   counts of its component trees.
 */
void
nautyextra_indnum_and_count_vpar (const int *vpar, int n,
                                  struct nautyextra_indnum_and_count *info)
{
  DEBUG(printf("nautyextra_indnum_vpar() n=%d", n);
        nautyextra_print_vpar(vpar,n));

  if (n <= 1) {
    info->indnum = n;
    info->count = 1;
    return;
  }
  int i;
  struct data {
    int                  num_children;
    int                  indnum_with;
    int                  indnum_without;
    NAUTYEXTRA_TYPE_ULL  count_with;
    NAUTYEXTRA_TYPE_ULL  count_without;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indnum_and_count_vpar");
#endif
  assert (n >= 0);

  if (n <= 1) {
    info->indnum = n;  /* 0 or 1 vertex set */
    info->count = 1;   /* single set, empty or 1 vertex */
    return;
  }

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].indnum_with   = 1;
    data[i].count_with    = 1;
    data[i].count_without = 1;
  }

  int                 total_indnum = 0;
  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d indnum %d,%d\n",
                 v, data[v].num_children,
                 data[v].indnum_with, data[v].indnum_without));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_indnum_with    = data[v].indnum_with;
      int c_indnum_without = data[v].indnum_without;
      NAUTYEXTRA_TYPE_ULL c_count_with    = data[v].count_with;
      NAUTYEXTRA_TYPE_ULL c_count_without = data[v].count_without;
      DEBUG(printf("consider v=%d indnum %d,%d counts %Lu,%Lu\n",
                   v, c_indnum_with, c_indnum_without,
                   c_count_with, c_count_without));
      assert (data[v].num_children == 0);

      /* Larger of c with or without.
         If with==without then count is both with + without.
      */
      int                 c_indnum_with_or_without;
      NAUTYEXTRA_TYPE_ULL c_count_with_or_without;
      if (c_indnum_with >= c_indnum_without) {
        c_indnum_with_or_without = c_indnum_with;
        if (c_indnum_with == c_indnum_without) {
          DEBUG(printf("  without v, child with=without, sum %Lu\n",
                       c_count_with+c_count_without));
          ADD_NO_OVERFLOW_ULL(c_count_with_or_without,
                              c_count_with, c_count_without);
        } else {
          DEBUG(printf("  without v, child with bigger\n"));
          c_count_with_or_without = c_count_with;
        }
      } else {
        DEBUG(printf("  without v, child without bigger\n"));
        c_indnum_with_or_without = c_indnum_without;
        c_count_with_or_without  = c_count_without;
      }

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  root\n"));
        total_indnum += c_indnum_with_or_without;
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_count_with_or_without);
        break;
      }

      /* with v must have child without */
      data[v].indnum_with += c_indnum_without;
      MUL_NO_OVERFLOW_ULL(data[v].count_with,
                          data[v].count_with, c_count_without);

      data[v].indnum_without += c_indnum_with_or_without;
      MUL_NO_OVERFLOW_ULL(data[v].count_without,
                          data[v].count_without, c_count_with_or_without);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d  %d,%d %Lu,%Lu\n",
                     v, data[v].indnum_with, data[v].indnum_without,
                     data[v].count_with, data[v].count_without));
        break;
      }
    }
  }

  info->indnum = total_indnum;
  info->count  = total_count;
}


/* ------------------------------------------------------------------------- */
/* Matchings */

/* An independent edge set, or matching, is a set of vertex pairs with an
   edge between each and no vertex in more than one pair.

   The matching number (here "matchnum") is the number of pairs in the
   largest matching of a graph.
*/

/* Return the matching number of vpar[1..n].
   vpar can be a forest, in which case the effect is sum matching numbers of
   its component trees.

   The approach is to find a leaf vertex and pair it with its attachment
   parent.  If an attachment is ever paired then pairing it with the leaf
   will be a set size >= what could be achieved by any other.  The pair is
   then reckoned as deleted from the tree and the process repeated.  The
   code works upwards from leaf vertices and pairs with parent as long as
   the parent is not already matched.
*/
int
nautyextra_matchnum_vpar (const int *vpar, int n)
{
  int i;
  struct data {
    int  num_children;
    int  matched;          /* 0 or 1 */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_matchnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_matchnum_vpar() "); nautyextra_print_vpar(vpar,n));

  if (n < 2) return 0;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++)
    data[vpar[i]].num_children++;

  int matchnum = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_matched = data[v].matched;
      assert (c_matched == 0 || c_matched == 1);
      DEBUG(printf("consider v=%d matched %d\n", v, c_matched));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) break;  /* root   */

      /* take this pair leaf c and parent v when both unmatched */
      int take_c_v = (1-c_matched) & (1-data[v].matched);
      assert (take_c_v == 0 || take_c_v == 1);
      matchnum += take_c_v;
      data[v].matched |= take_c_v;

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;  /* non-leaf */
    }
  }
  return matchnum;
}

/* Set *info to the matching number and count of matchings of this size in
   tree vpar[1..n].

   c_big       = MAX(c_with, c_without)
   c_big_count = c_with_count or c_without_count per c_big

   v_without       = sum c_big
   v_without_count = prod c_big_count

   v_with = 1 + max[ c_without + others sum c_big ]
          = 1 + v_without - min c_with_is_bigger         one ensured without
   v_with_count = sum  c_without * others prod c_big_count  of those max
   when new low c_with_is_bigger for v_without_sub_with_plus_1,
   reset v_with_count=0
   when equal low c_with_is_bigger,
   add c_without * v_without_count before its update

   v_without_sub_with_plus_1
   = v_without - v_with + 1 = min c_with_is_bigger

   c_with = c_without - c_without_sub_with_plus_1 + 1

             1
           /   \        c=2 h=3 matching number 2
         2       3
        / \     / \
       4   5   6   7
 */
void
nautyextra_matchnum_and_count_vpar (const int *vpar, int n,
                                    struct nautyextra_matchnum_and_count *info)
{
  DEBUG(printf("nautyextra_matchnum_and_count_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int                  with;
    int                  without_is_smaller;
    NAUTYEXTRA_TYPE_ULL  with_count;
    NAUTYEXTRA_TYPE_ULL  without_count;
    int                  num_children;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_matchnum_and_count_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].without_count = 1;  /* others start all 0 */
  }

  int                  ret = 0;
  NAUTYEXTRA_TYPE_ULL  ret_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      DEBUG(printf("consider v=%d with %d[%Lu] smaller %d [%Lu] parent %d\n",
                   v, data[v].with, data[v].with_count,
                   data[v].without_is_smaller,
                   data[v].without_count, vpar[v]));
      assert (data[v].num_children == 0);

      NAUTYEXTRA_TYPE_ULL v_max_count = data[v].with_count;
      if (! data[v].without_is_smaller)
        ADD_NO_OVERFLOW_ULL(v_max_count, v_max_count, data[v].without_count);
      DEBUG(printf(" v_max_count %Lu\n", v_max_count));

      data[v].num_children = -1;
      int p = vpar[v];     /* to parent */
      if (! p) {
        ret += data[v].with;
        MUL_NO_OVERFLOW_ULL(ret_count, ret_count, v_max_count);
        DEBUG(printf(" root add %d [%Lu] to ret now %d [%Lu]\n",
                     data[v].with,v_max_count, ret,ret_count));
        break;
      }

      assert(data[v].without_is_smaller==0 || data[v].without_is_smaller==1);
      assert(data[p].without_is_smaller==0 || data[p].without_is_smaller==1);

      int s = data[p].without_is_smaller + data[v].without_is_smaller;
      assert (s==0 || s==1 || s==2);

      DEBUG(printf(" parent %d  with %d [%Lu] smaller %d [%Lu]\n",
                   p, data[p].with, data[p].with_count,
                   data[p].without_is_smaller,
                   data[p].without_count));

      data[p].with += data[v].with + (s==0);
      if (s==0) { data[p].without_is_smaller = 1; }

      /* with_count[p] = if(s>=1, with_count[p] * v_max_count)
         .             + if(s<=1, without_count[p] * without_count[v]);
      */
      NAUTYEXTRA_TYPE_ULL p_with_count = 0;
      if (s >= 1)
        MUL_NO_OVERFLOW_ULL(p_with_count, data[p].with_count, v_max_count);
      if (s <= 1) {
        NAUTYEXTRA_TYPE_ULL  temp;
        MUL_NO_OVERFLOW_ULL(temp, data[p].without_count, data[v].without_count);
        ADD_NO_OVERFLOW_ULL(p_with_count, p_with_count, temp);
      }
      data[p].with_count = p_with_count;

      /* without_count[p] = prod v_max_count */
      MUL_NO_OVERFLOW_ULL(data[p].without_count,
                          data[p].without_count, v_max_count);

      assert(data[p].num_children > 0);
      if (--data[p].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }

      v = p;
    }
  }

  DEBUG(printf(" return %d count %Lu\n",
               ret, (unsigned long long) ret_count));
  info->matchnum = ret;
  info->count = ret_count;
}


/* ------------------------------------------------------------------------- */
/* Perfect Matching */

/* A perfect matching is one where all vertices in the graph are paired in a
   matching.  This can exist only for an even number of vertices.
*/

/* Return TRUE if tree vpar[1..n] has a perfect matching.
   Return FALSE if not.

   The approach is the same as nautyextra_matchnum_vpar() above, but if any
   leaf is found which cannot be matched, because its parent is already
   matched, then immediately return FALSE.
*/
int
nautyextra_has_perfect_matching_vpar (const int *vpar, int n)
{
  assert (n >= 0);
  if (n&1) return 0;

  int i;
  struct data {
    int  num_children;
    int  paired;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_has_perfect_matching_vpar");
#endif

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }
    for (;;) {
      int c_paired = data[v].paired;
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        if (! c_paired)
          return FALSE;
        break;
      }

      /* If child not already paired then do so with parent.
         If parent already paired (with some other child) then no perfect
         matching possible.
      */
      if (! c_paired) {
        if (data[v].paired++)
          return FALSE;
      }

      assert(data[v].num_children > 0);
      if (--data[v].num_children)
        break;  /* stop at non-leaf */
    }
  }
  return TRUE;
}


/* ------------------------------------------------------------------------- */
/* Dominating Sets */

/* A dominating set (here "domset") is a set of vertices where all vertices
   of the graph are either in the set or a neighbour of some vertex in the
   set.
*/

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if ss is a dominating set in sg.  Return FALSE if not.
*/
int
nautyextra_is_domset_sg (const sparsegraph *sg, const set *ss)
{
  /* cf setnbhd() getting all neighbours of all in a set */

  DEBUG(printf("nautyextra_is_domset_sg() n=%d  %s\n",
               sg->nv, nautyextra_set_str(ss,sg->nv)));
  int n = sg->nv;

  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int v;
  for (v=0; v<n; v++) {
    if (ISELEMENT(ss,v)) continue;  /* v in ss, good */
    const int *eptr = ee + vv[v];
    int degree = dd[v];
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      if (ISELEMENT(ss,to)) goto NEXT_V;  /* neighbour "to" in ss, good */
    }
    DEBUG(printf("  v=%d not dominated\n", v));
    return FALSE;  /* v is not dominated */
  NEXT_V: ;
  }
  DEBUG(printf("  is domset\n"));
  return TRUE;
}

/* Return mask 0xFF..FF if high bit element 0 of ss is set, or return 0 if
   high bit not set.
*/
static inline setword
highbit_to_mask (setword ss)
{
  return - ((ss & NAUTYEXTRA_SETWORD_HIGHBIT) >> (WORDSIZE-1));
}

/* g,m=1,n is an undirected graph.
   ss is a set of vertices 0 to n-1.
   Return TRUE if ss is a dominating set in g.  Return FALSE if not.
*/
int
nautyextra_is_domset1 (const graph *g, int n, setword ss)
{
#if WORDSIZE == 32
  assert (highbit_to_mask(0xFFFFFFFF) == 0xFFFFFFFF);
  assert (highbit_to_mask(0x7FFFFFFF) == 0);
#endif

  DEBUG(printf("nautyextra_is_domset1() ss=%s (%X)\n",
               nautyextra_set_str(&ss,n), ss));
  assert ((ss & BITMASK(n)) == 0);   /* no stray bits in ss */

  setword dominated = ss;
  int v;
  for (v=0; v<n; v++) {
    dominated |= g[v] & highbit_to_mask(ss);
    DEBUG(printf(" v=%d g %s", v, nautyextra_set_str(g+v,n));
          printf(" ss %-*s", n, nautyextra_set_str(&ss,n-v));
          setword mask = highbit_to_mask(ss);
          printf(" mask %s", nautyextra_set_str(&mask,n));
          printf(" to dominated %s\n", nautyextra_set_str(&dominated,n)));
    ss <<= 1;
  }
  DEBUG(printf(" final dominated %s (%X want %X)\n",
               nautyextra_set_str(&dominated,n), dominated, ALLMASK(n)));
  return dominated == ALLMASK(n);
}

/* sg is an undirected sparsegraph.
   Return the number of dominating sets in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   few vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE many vertices.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_domsets_count_sg (const sparsegraph *sg)
{
  setword count = 0;
  setword ss;
  int n = sg->nv;

  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_domsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_domset_sg (sg,&ss));
  return count;
}

/* Return the number of dominating sets in tree vpar[1..n].
   abort() on overflow.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_domsets_count_vpar (const int *vpar, int n)
{
  /* The approach is to establish 3 values at each vertex, counting how many
     sets with or without v, and without v and v undominated.

     .    v_with          = count sets with v included
     .    v_without       = count sets without v, and v either dom or undom
     .    v_without_undom = count sets without v, and v undominated
     .    all counts are with the children of v (if any) dominated

     Sets without v but where v is dominated, by some child, are difference

     .    v_without_dom = v_without - v_without_undom

     Total domsets at the root is those plus v_with,

     .    v_dom = v_with + v_without_dom

     The counts at vertex v follow from products of corresponding child
     counts,

     .    v_with          = prod(c_any)                         v
     .    v_without       = prod(c_dom)                      /  |  \
     .    v_without_undom = prod(c_without_dom)            c1  c2  ...
     .    where c_any = c_with + c_without

     For v_with, can have undominated children, since v dominates them.
     For v_without, can have all the child domsets.  They might or might not
     dominate v.
     For v_without_undom, only child sets without c (but c itself dominated
     from below since there will be no v to do it).

     The products start as all 1 at all vertices.  For a single vertex tree
     this gives v_dom = 1 + (1 - 1) = 1 as desired.
  */
  assert (n >= 0);
  if (n <= 1) {
    /* empty set dominates empty graph, or 1 vertex dominates 1 vertex */
    return 1;
  }

  int i;
  struct data {
    int                 num_children;
    NAUTYEXTRA_TYPE_ULL  with;
    NAUTYEXTRA_TYPE_ULL  without;
    NAUTYEXTRA_TYPE_ULL  without_undom;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_domsets_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domsets_count_vpar()");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with = 1;
    data[i].without = 1;
    data[i].without_undom = 1;
  }

  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      NAUTYEXTRA_TYPE_ULL  c_with          = data[v].with;
      NAUTYEXTRA_TYPE_ULL  c_without       = data[v].without;
      NAUTYEXTRA_TYPE_ULL  c_without_undom = data[v].without_undom;
      DEBUG(printf("consider v=%d child %Lu,%Lu,%Lu\n",
                   v, c_with, c_without, c_without_undom));
      assert (data[v].num_children == 0);

      assert(c_without >= c_without_undom);
      NAUTYEXTRA_TYPE_ULL  c_without_dom  = c_without - c_without_undom;

      NAUTYEXTRA_TYPE_ULL  c_dom;  /* c_dom = count sets with c dominated */
      ADD_NO_OVERFLOW_ULL(c_dom, c_with, c_without_dom);

      data[v].num_children = -1;  /* drop this leaf */
      v = vpar[v];                /* to parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_dom);
        break;
      }

      NAUTYEXTRA_TYPE_ULL  c_any;
      ADD_NO_OVERFLOW_ULL(c_any, c_with, c_without);

      DEBUG(printf("  any %Lu dom %Lu, with %Lu, without %Lu = dom %Lu + undom %Lu\n",
                   c_any, c_dom, c_with, c_without, c_without_dom, c_without_undom));

      /* with = prod(c_any) */
      MUL_NO_OVERFLOW_ULL(data[v].with,
                          data[v].with, c_any);

      /* without = prod(c_dom) */
      MUL_NO_OVERFLOW_ULL(data[v].without,
                          data[v].without, c_dom);

      /* undom = prod(c_without_dom) */
      MUL_NO_OVERFLOW_ULL(data[v].without_undom,
                          data[v].without_undom, c_without_dom);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf\n"));
        break;
      }
    }
  }

  DEBUG(printf("  total_count %Lu\n", total_count));
  return total_count;
}


/* ------------------------------------------------------------------------- */
/* Minimal Dominating Sets */

/* A minimal dominating set is a set from which it's not possible to remove
   a vertex and still be a dominating set (here "minimal_domset").
*/

/* g,m=1,n is an undirected graph.
   ss is a set of vertices 0 to n-1.
   Return TRUE if ss is a minimal dominating set in g.  Return FALSE if not.
*/
int
nautyextra_is_minimal_domset1 (const graph *g, int n, setword ss)
{
  /* ENHANCE-ME: Maybe look for >=2 dom in second progressively for early
     exit.  Still need to look at the end in case a non-minimal seen before
     some surrounding minimals show its non-minimality.
   */
  DEBUG(printf("nautyextra_is_minimal_domset1() ss=%s (%X)\n",
               nautyextra_set_str(&ss,n), ss));
  assert ((ss & BITMASK(n)) == 0);   /* no stray bits in ss */

  setword second    = 0;
  setword dominated = ss;
  setword s    = ss;
  int v;
  for (v=0; v<n; v++, s<<=1) {
    setword this = g[v] & highbit_to_mask(s);
    DEBUG(printf(" v=%d this %s", v, nautyextra_set_str(&this,n));
          printf(" dominated %s", nautyextra_set_str(&dominated,n));
          printf(" second %s\n", nautyextra_set_str(&second,n)));
    second |= dominated & this;
    dominated |= this;
  }
  DEBUG(printf(" final dominated %s (%X want %X)",
               nautyextra_set_str(&dominated,n), dominated, ALLMASK(n));
        printf(" second %s\n", nautyextra_set_str(&second,n)));
  if (dominated != ALLMASK(n)) return FALSE;

  setword vs = NAUTYEXTRA_SETWORD_HIGHBIT;
  second = ~second;
  for (v=0; v<n; v++, vs>>=1) {
    if ((ss & vs) && !(second & (vs|g[v]))) {
      DEBUG(printf(" v=%d and neighbours all >=2 dominations, not minimal\n",
                   v));
      return FALSE;
    }
  }
  return TRUE;
}

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if set ss is a minimal dominating set in sg.
   Return FALSE if not.
*/
int
nautyextra_is_minimal_domset_sg (const sparsegraph *sg, const set *ss)
{
  return nautyextra_is_domset_sg(sg,ss)
    && nautyextra_domset_is_minimal_sg (sg,ss);
}

/* sg is an undirected sparsegraph.
   ss is a set of vertices 0 to sg->nv-1 which dominates some of sg.
   Return TRUE if ss is minimal, meaning that no vertex could be removed
   from ss and still dominate the same vertices.
   Return FALSE if not.

   A vertex which could be removed has itself and all neighbours dominated
   by others in ss.  The approach is to count how many dominations there are
   at each vertex then search for a v in ss >=2 dominations of itself and
   neighbours.  The search is for >=2 because the count includes v's own
   dominating.
*/
int
nautyextra_domset_is_minimal_sg (const sparsegraph *sg, const set *ss)
{
  int n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  if (n==0) return TRUE;

  int m = SETWORDSNEEDED(n);
  int v;
#if MAXN
  int dominated[MAXN];
  assert (n <= MAXN);
#else
  WORKSPACE(int,dominated,n,"nautyextra_domset_is_minimal_sg");
#endif

  DEBUG(printf("nautyextra_domset_is_minimal_sg() n=%d m=%d %s\n",
               n, m, nautyextra_set_str(ss,n)));

  memset(dominated, '\0', n*sizeof(dominated[0]));

  /* dominated[v] is count of how many vertices of ss dominate it
     (including v itself if v is in ss) */
  for (v = -1; (v = nextelement((set*)ss,m,v)) >= 0; ) {
    assert(v < n);
    dominated[v]++;
    int epos   = vv[v];
    assert (0 <= epos && epos <= sg->elen);
    const int *eptr = ee + epos;
    int degree = dd[v];
    assert (epos + degree <= sg->elen);
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      dominated[to]++;
    }
  }

  DEBUG(printf("  dominated counts ");
        for (v = 0; v < n; v++) printf(" %d",dominated[v]);
        printf("\n"));

  /* Search for a v in ss which is itself and all neighbours dominated >=2
     times, meaning that it and neighbours are covered by v but also by some
     others of ss, allowing v to be removed.
  */
  for (v = -1; (v = nextelement((set*)ss,m,v)) >= 0; ) {
    DEBUG(printf("  v=%d dom %d\n", v, dominated[v]));
    if (dominated[v] < 2) goto NEXT_SS;
    const int *eptr = ee + vv[v];
    int degree = dd[v];
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      DEBUG(printf("    neighbour %d dom %d\n", to, dominated[to]));
      if (dominated[to] < 2) goto NEXT_SS;
    }
    /* v has itself and neighbours dominated >=2 times, so it is overlapped
       by other members of ss
    */
    DEBUG(printf("  v=%d overlapped\n", v));
    return FALSE;

  NEXT_SS:
    ;
  }
  DEBUG(printf("  is minimal\n"));
  return TRUE;
}

/* Return the number of minimal dominating sets in g,m=1,n.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_minimal_domsets_count1 (const graph *g, int n)
{
  /* ENHANCE-ME: Maybe maintain the dominated/second bit sets of
     nautyextra_is_minimal_domset1() and update them for Gray code changes
     to the set ss.
  */
  assert (n >= 0);
  assert (n <= WORDSIZE);

  if (n == 0) return 1;   /* n=0 empty set is only domset */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_minimal_domset1(g,n,ss));
  return count;
}

/* Return the number of minimal dominating sets in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   few vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE vertices.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_minimal_domsets_count_sg (const sparsegraph *sg)
{
  /* ENHANCE-ME: Leaf vertices and their attachments are mutually exclusive.
     Can prune the loop if find a non-minimal vertex in the first few.
  */
  int n = sg->nv;
  assert (n >= 0);

  if (n == 0) return 1;   /* n=0 empty set is only domset */
  DEBUG(printf("nautyextra_minimal_domsets_count_sg()\n");
        nautyextra_print_sg(sg));
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_minimal_domsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_minimal_domset_sg(sg,&ss));
  return count;
}

/* Return the number of minimal dominating sets in tree vpar[1..n].
   */
NAUTYEXTRA_TYPE_ULL
nautyextra_minimal_domsets_count_vpar (const int *vpar, int n)
{
  /* The approach is to accumulate 6 quantities at each vertex, based on
     corresponding values determined in each of its children.

     .   with
     .   with_notreq
     .   with_notmin_notreq
     .   without_notsole
     .   without_dom_sole
     .   without_undom

     "with" is count of sets which include the given vertex v.  The children
     and all below them are minimal in the sense that anything removed from
     the set would not be dominating.  But vertex v itself might or might
     not be minimal.

     "min" is when v is minimal (its removal would not be a dominating set).
     "notmin" is when v is definitely not minimal, so it could be removed
     and still be a dominating set.  (That set would then be minimal, since
     the "with" sets are minimal in the children and below.)

     "req" is when v is required in the set for the purpose of dominating
     one or more of the child vertices (those being otherwise undominated).
     "notreq" is when v is not required for that purpose, ie. all children
     are dominated by some other vertex in the set too.

     The combinations in "with" are then

     .                 req          notreq
     .                 ----         ------
     .    min    |  with_min_req   with_min_notreq        row = with_min
     .   notmin  |      0          with_notmin_notreq
     .
     .                             column = with_notreq

     There are no "notmin_req" sets, since if v is required for dominating
     something then it can't be removed hence "min" minimal.

     The quantities kept are "with" which is the whole table, "with_notreq"
     which is just the right column, and "with_min_notreq" which is just the
     top right.


     "without" is count of sets which do not include v.  The children and
     everything below are dominated and minimal, but v itself might or might
     not be dominated.

     "dom" is when v is dominated (some child is in the set).
     "undom" is when v is undominated (no child in the set).

     "sole" is sets where v is dominated by one child and this must be the
     sole domination of v because any further domination of v would make
     that child not minimal.  "notsole" is where can be further dominated
     without causing its children to be not minimal.

     The combinations in "without" are then

     .               sole               notsole
     .               ----               -------
     .   dom    without_dom_sole    without_dom_notsole    row = without_dom
     .   undom        0             without_undom
     .
     .                              column = without_notsole

     There are no "undom_sole" combinations.  The "sole" is when v dominated
     by a child.

     The quantities kept are "without_notsole" which is the right column and
     the individual "without_dom_sole" and "without_undom".  top left and
     bottom right respectively,


     The 6 quantities at v are determined by the following products of child
     quantities.

     v_with
     = prod(c_with_req + c_without_dom_notsole + c_without_undom)
     = prod(c_with_req +               c_without_notsole        )

     v_with_notreq
     = prod(c_with_req + c_without_dom_notsole                  )

     v_with_min_notreq
     = prod(             c_without_dom_notsole                  )

     v_without_notsole = prod(c_with_min + c_without_dom)

     v_without_undom   = prod(             c_without_dom)

     v_without_dom_sole
     = sum  / c_with_notmin_notreq * prod(c_without_dom)  \
     .  c   \                        other c              /

     "v_with" has children c_with_req.  It does not have c_with_notreq since
     if a child is not required below then the presence of v above would
     make the child not minimal.  Its children without are all the notsole
     so that they can take further domination by v.

     "v_with_notreq" is a subset of v_with.  Any c_without_undom makes v
     required for dominating a child, hence those omitted is notreq.
     v still might or might not be be minimal.

     "v_with_min_notreq" is a further subset.  It make v minimal by having
     only children without, so v is the only domination of v.  The children
     are type c_without_dom_notsole since the further domination of them by
     v must be allowed.

     "v_without_notsole" is c_with_min and c_without_dom.  Since no v to
     dominate the children must be c_without_dom.  The children with are the
     minimals c_with_min so that further domination of v is allowed for
     notsole.  Children which are c_with_notmin_notreq would not allow that.

     "v_without_undom" is a subset of v_without_notsole.  It leaves v
     undominated by taking only c_without_dom.  Again the children are
     already dominated since there is no v to dominate them.

     "v_without_dom_sole" is the v "sole" combinations.  There is to be
     exactly 1 child c_with_notmin_notreq dominating v and other children
     c_without_dom.  The sum is over which child is the c_with_notmin_notreq
     and the product is c_without_dom of the other children.

     v_without_dom_sole is accumulated by multiplying new c_without_dom for
     this child as another of the "others" in the sum, then add
     c_with_notmin_notreq times the others product.  That others product is
     "v_without_undom__so_far" which is product c_without_dom before the
     current child is multiplied in.

     v_without_dom_sole = v_without_dom_sole * c_without_dom
     .                  + v_without_undom__so_far * c_with_notmin_notreq

     The "c" terms used above include the following formed by difference from
     the 6 quantities accumulated in the children,

     c_with_req = c_with_min_req  = c_with - c_with_notreq
     c_with_notmin_notreq         = c_with_notreq - c_with_min_notreq
     c_with_min                   = c_with - c_with_notmin_notreq

     c_without_dom_notsole = c_without_notsole - c_without_undom
     c_without_dom         = c_without_dom_sole + c_without_dom_notsole

     Vertices begin with all products = 1 and sum v_without_dom_sole = 0.
     This includes leaf vertices.  These initial values at a leaf give counts
     for them which are per the definitions.

     The final count of minimal dominating sets is minimal with and
     dominated without,

     v_mindomsets = v_with_min + v_without_dom

     Notice this is used to accumulate into v_without_notsole above, so is
     calculated for use either as that or as the final return.

     Expanded out in terms of the 6 products the final result is

     v_mindomsets =   v_with            - v_with_notreq   + v_with_min_notreq
     .              + v_without_notsole - v_without_undom + v_without_dom_sole
  */

  if (n == 0) return 1;  /* empty set dominates empty graph */

  int i;
  struct data {
    int                  num_children;
    NAUTYEXTRA_TYPE_ULL  with;
    NAUTYEXTRA_TYPE_ULL  with_notreq;
    NAUTYEXTRA_TYPE_ULL  with_min_notreq;
    NAUTYEXTRA_TYPE_ULL  without_dom_sole;
    NAUTYEXTRA_TYPE_ULL  without_notsole;
    NAUTYEXTRA_TYPE_ULL  without_undom;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_minimal_domsets_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_minimal_domsets_count_vpar()");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with               = 1;
    data[i].with_notreq        = 1;
    data[i].with_min_notreq    = 1;
    /* data[i].without_dom_sole   = 0; */  /* already 0 */
    data[i].without_notsole    = 1;
    data[i].without_undom      = 1;
  }

  NAUTYEXTRA_TYPE_ULL total_count = 1;  /* for n=0 vertices */
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children) { continue; }

    for (;;) {
      NAUTYEXTRA_TYPE_ULL  c_with             = data[v].with;
      NAUTYEXTRA_TYPE_ULL  c_with_notreq      = data[v].with_notreq;
      NAUTYEXTRA_TYPE_ULL  c_with_min_notreq  = data[v].with_min_notreq;
      NAUTYEXTRA_TYPE_ULL  c_without_notsole  = data[v].without_notsole;
      NAUTYEXTRA_TYPE_ULL  c_without_dom_sole = data[v].without_dom_sole;
      NAUTYEXTRA_TYPE_ULL  c_without_undom    = data[v].without_undom;

      assert (c_with_notreq >= c_with_min_notreq);
      NAUTYEXTRA_TYPE_ULL  c_with_notmin_notreq
        = c_with_notreq - c_with_min_notreq;

      /* c_with_min = c_with - (c_with_notreq - c_with_min_notreq)
         .          = c_with - c_with_notmin_notreq
      */
      assert (c_with >= c_with_notmin_notreq);
      NAUTYEXTRA_TYPE_ULL  c_with_min = c_with - c_with_notmin_notreq;

      assert (c_without_notsole >= c_without_undom);
      NAUTYEXTRA_TYPE_ULL  c_without_dom_notsole
        = c_without_notsole - c_without_undom;

      NAUTYEXTRA_TYPE_ULL  c_without_dom;
      ADD_NO_OVERFLOW_ULL(c_without_dom,
                          c_without_dom_sole, c_without_dom_notsole);

      NAUTYEXTRA_TYPE_ULL c_mindom;
      ADD_NO_OVERFLOW_ULL(c_mindom, c_with_min, c_without_dom);

      DEBUG(printf("consider v=%d\n", v));
      assert (data[v].num_children == 0);

      /* DEBUG(printf("  with %Lu, without %Lu = dom %Lu + undom %Lu\n", */
      /*              c_with, c_without, c_without_dom, c_without_undom)); */

      data[v].num_children = -1;  /* drop this leaf */
      v = vpar[v];                /* to parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_mindom);
        break;
      }

      /* c_with_req = c_with_min_req, the only type of "with req" */
      assert (c_with >= c_with_notreq);
      NAUTYEXTRA_TYPE_ULL  c_with_req = c_with - c_with_notreq;

      /* with = prod(c_with_req + c_without_dom_notsole + c_without_undom)
         .    = prod(c_with_req +             c_without_notsole)
      */
      NAUTYEXTRA_TYPE_ULL  c_with_req__plus__without_notsole;
      ADD_NO_OVERFLOW_ULL(c_with_req__plus__without_notsole,
                          c_with_req, c_without_notsole);
      MUL_NO_OVERFLOW_ULL(data[v].with,
                          data[v].with, c_with_req__plus__without_notsole);

      /* with_notreq = prod(c_with_req + c_without_dom_notsole)
       */
      assert(c_with_req__plus__without_notsole >= c_without_undom);
      NAUTYEXTRA_TYPE_ULL  c_with_req__plus__without_dom_notsole
        = c_with_req__plus__without_notsole - c_without_undom;
      assert (c_with_req__plus__without_dom_notsole
              == c_with_req + c_without_dom_notsole);
      MUL_NO_OVERFLOW_ULL(data[v].with_notreq,
                          data[v].with_notreq,
                          c_with_req__plus__without_dom_notsole);

      /* with_min_notreq = prod(c_without_dom_notsole) */
      MUL_NO_OVERFLOW_ULL(data[v].with_min_notreq,
                          data[v].with_min_notreq, c_without_dom_notsole);

      /* v_without_dom_sole
         = sum one of c_with_notmin_notreq * prod(c_without_dom)
      */
      NAUTYEXTRA_TYPE_ULL  v_without_dom_sole;
      MUL_NO_OVERFLOW_ULL(v_without_dom_sole,
                          data[v].without_dom_sole, c_without_dom);
      NAUTYEXTRA_TYPE_ULL v_without_undom__so_far
        = data[v].without_undom; /* before multiplying in new product term */
      NAUTYEXTRA_TYPE_ULL without_dom_sole__add;
      MUL_NO_OVERFLOW_ULL(without_dom_sole__add,
                          v_without_undom__so_far, c_with_notmin_notreq);
      ADD_NO_OVERFLOW_ULL(data[v].without_dom_sole,
                          v_without_dom_sole, without_dom_sole__add);

      /* without_notsole = prod(c_with_min + c_without_dom)
         .               = prod(c_mindom)
      */
      MUL_NO_OVERFLOW_ULL(data[v].without_notsole,
                          data[v].without_notsole, c_mindom);

      /* v_without_undom = prod(c_without_dom) */
      MUL_NO_OVERFLOW_ULL(data[v].without_undom,
                          v_without_undom__so_far, c_without_dom);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  DEBUG(printf("  total_count %Lu\n", total_count));
  return total_count;
}


/* ------------------------------------------------------------------------- */
/* Domination Number */

/* The domination number is the size of the smallest dominating set in a
   graph (here "domnum").
*/

/* Set *info to the domination number and count of sets of this size in g,1,n.
*/
void
nautyextra_domnum_and_count1 (const graph *g, int n,
                              struct nautyextra_domnum_and_count *info)
{
  DEBUG(printf("nautyextra_domnum_and_count() n=%d\n", n));
  assert (n >= 0);
  if (n == 0) {
    DEBUG(printf(" n=0 special case\n"));
    info->domnum = 0; /* empty set of vertices */
    info->count = 1;  /* 1 way */
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_domnum_and_count() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int domnum = n+1;  /* to be bettered by n vertices or fewer */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     DEBUG(printf(" consider set %s\n", nautyextra_set_str(&ss,n)));
     if (nautyextra_is_domset1(g,n,ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < domnum) {
         DEBUG(printf(" new low %d\n", ss_size));
         domnum = ss_size;
         count = 1;
       } else if (ss_size == domnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end domnum %d count %u\n", domnum, count));
  info->domnum = domnum;
  info->count = count;
}


/* Set *info to the domination number and count of sets of this size in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   small number of vertices, and currently only uses 1-word sets, so must
   have sg->nv <= WORDSIZE many vertices.
*/
void
nautyextra_domnum_and_count_sg (const sparsegraph *sg,
                                struct nautyextra_domnum_and_count *info)
{
  DEBUG(printf("nautyextra_domnum_and_count_sg() n=%d\n", sg->nv));

  /* ENHANCE-ME: Per Cockayne, Goodman, and Hedetniemi, at any leaf vertices
     should take the attachment and exclude the leaf.  The leaf can be
     dominated by taking either that leaf or its attachment.  Taking the
     attachment gives a set size <= taking the leaf.
  */

  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    DEBUG(printf(" n=0 special case\n"));
    info->domnum = 0; /* empty set of vertices */
    info->count = 1;  /* 1 way */
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_domnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int domnum = n+1;  /* to be bettered by n vertices or fewer */
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     DEBUG(printf(" consider set %s\n", nautyextra_set_str(&ss,n)));
     if (nautyextra_is_domset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < domnum) {
         DEBUG(printf(" new low %d\n", ss_size));
         domnum = ss_size;
         count = 1;
       } else if (ss_size == domnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end domnum %d count %u\n", domnum, count));
  info->domnum = domnum;
  info->count = count;
}

/* Return the domination number of tree vpar[1..n].
   vpar can be a forest, in which case the result is sum of domination
   numbers of its component trees.

   The code follows

       Cockayne, Goodman, and Hedetniemi, "A Linear Algorithm for the
       Domination Number of a Tree", Information Processing Letters,
       volume 4, number 2, November 1975, pages 41-44.

   At a leaf must have either it or its attachment vertex in the dominating
   set.  The two are both 1 vertex, but choosing the attachment dominates
   more than the leaf, so makes a dominating set <= taking the leaf.  The
   leaf can then be discarded.

   Any leaf which is already dominated by a previously taken vertex can be
   discarded likewise.  A leaf which is required in the dominating set then
   makes its parent dominated.  If the parent is itself required then it
   remains required.
 */
int
nautyextra_domnum_vpar (const int *vpar, int n)
{
  int domnum = 0;
  int i;
  struct data {
    int  num_children;
    enum mtype { MTYPE_UNDOMINATED = 0,
                 MTYPE_DOMINATED, /* absent from set but dominated by another */
                 MTYPE_REQUIRED
    } mtype;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_domnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domnum_vpar() ");
        nautyextra_print_vpar(vpar,n));

  if (n==0) return 0;
  if (n==1) return 1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      enum mtype  c_mtype = data[v].mtype;
      DEBUG(printf("consider v=%d (parent %d) mtype %d  (domnum %d)\n",
                   v, vpar[v], c_mtype, domnum));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        domnum += (c_mtype != MTYPE_DOMINATED);
        goto END;
      }

      switch (c_mtype) {
      case MTYPE_DOMINATED:
        /* already dominated, discard */
        break;
      case MTYPE_UNDOMINATED:
        /* not dominated, must have parent */
        data[v].mtype = MTYPE_REQUIRED;
        break;
      case MTYPE_REQUIRED:
        /* required, so parent can raise from MTYPE_UNDOMINATED to MTYPE_DOMINATED
           (if parent already MTYPE_REQUIRED then it stays that way) */
        if (data[v].mtype == MTYPE_UNDOMINATED)
          data[v].mtype = MTYPE_DOMINATED;
        domnum++;
        break;
      default:
        assert(0);
      }

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d   %d\n", v, data[v].mtype));
        break;
      }
    }
  }

 END:
  DEBUG(printf("  final %d\n", domnum));
  return domnum;
}

/* Set *info to the domination number and count of sets of this size in
   tree vpar[1..n].
   vpar can be a forest, in which case the effect is domination number sum
   of component trees, and count product of component trees.


   v_without = sum c_dom              v dom or undom
   v_without_count = prod c_dom

   c_any = min(c_with, c_without)
   v_with_count = prod c_any

   c_dom = min(c_with, c_without_dom)
   v_without_dom = min(c_with + sum others c_dom)

   prospective = c_with + sum others c_dom (= v_without)
   cf            c_dom  + v_without_dom so far

   prospective_count = c_with_count * prod others c_dom_count (=v_without_count)

   smaller  v_without_dom_count = prospective_count
   bigger   v_without_dom_count *= c_dom_count
   equal    v_without_dom_count *= c_without_dom_count
   .        v_without_dom_count += prospective_count

 */
void
nautyextra_domnum_and_count_vpar (const int *vpar, int n,
                                  struct nautyextra_domnum_and_count *info)
{
  DEBUG(printf("nautyextra_domnum_and_count_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  int i;
  struct data {
    int                  num_children;
    int                  with;
    int                  without;
    int                  without_dom;
    NAUTYEXTRA_TYPE_ULL  with_count;
    NAUTYEXTRA_TYPE_ULL  without_count;
    NAUTYEXTRA_TYPE_ULL  without_dom_count;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_domnum_and_count_vpar");
#endif

  if (n == 1) {
    info->domnum = n;
    info->count = 1;
    return;
  }

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with                = 1;
    data[i].without_dom         = n+1;
    data[i].with_count          = 1;
    data[i].without_count       = 1;
  }

  int total = 0;
  NAUTYEXTRA_TYPE_ULL  total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_with                           = data[v].with;
      int c_without                        = data[v].without;
      int c_without_dom                    = data[v].without_dom;
      NAUTYEXTRA_TYPE_ULL  c_with_count          = data[v].with_count;
      NAUTYEXTRA_TYPE_ULL  c_without_count       = data[v].without_count;
      NAUTYEXTRA_TYPE_ULL  c_without_dom_count   = data[v].without_dom_count;

      DEBUG(printf("consider v=%d with %d [%Lu] without %d [%Lu] without_dom %d [%Lu]   parent %d\n",
                   v, c_with, c_with_count,
                   c_without, c_without_count,
                   c_without_dom, c_without_dom_count,
                   vpar[v]));
      assert (data[v].num_children == 0);

      /* dom = min(with, without_dom) */
      int                 c_dom       = c_with;
      NAUTYEXTRA_TYPE_ULL c_dom_count = c_with_count;
      if (c_without_dom < c_dom) {
        c_dom       = c_without_dom;
        c_dom_count = c_without_dom_count;
      } else if (c_without_dom == c_dom) {
        ADD_NO_OVERFLOW_ULL(c_dom_count, c_dom_count, c_without_dom_count);
      }
      DEBUG(printf(" without_dom %d [%Lu]   dom %d [%Lu]\n",
                   c_without_dom, c_without_dom_count,
                   c_dom, c_dom_count));

      data[v].num_children = -1;
      v = vpar[v];     /* to parent */
      if (! v) {
        total += c_dom;
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_dom_count);
        DEBUG(printf(" root add %d [%Lu] to total now %d [%Lu]\n",
                     c_dom,c_dom_count, total,total_count));
        break;
      }

      /* any = min(with, without) */
      int                 c_any       = c_with;
      NAUTYEXTRA_TYPE_ULL c_any_count = c_with_count;
      assert (c_without_count > 0);
      if (c_without < c_any) {
        c_any       = c_without;
        c_any_count = c_without_count;
      } else if (c_without == c_any) {
        ADD_NO_OVERFLOW_ULL(c_any_count, c_any_count, c_without_count);
      }
      DEBUG(printf(" any %d [%Lu]\n", c_any, c_any_count));

      DEBUG(printf(" parent %d  with %d [%Lu] without %d [%Lu] without_dom %d [%Lu]\n",
                   v, data[v].with, data[v].with_count,
                   data[v].without, data[v].without_count,
                   data[v].without_dom, data[v].without_dom_count));

      /* v_with       = 1 + sum c_any
         v_with_count = prod c_any_count */
      data[v].with += c_any;
      MUL_NO_OVERFLOW_ULL(data[v].with_count,
                          data[v].with_count, c_any_count);

      int v_without                       = data[v].without;
      NAUTYEXTRA_TYPE_ULL v_without_count = data[v].without_count;

      /* v_without_dom   = min (>= one c_with, others c_dom)
         prospective = c_with + (others c_dom = v_without before this child)
         vs
         v_without_dom += c_without_dom
      */
      int                 v_without_dom       = data[v].without_dom;
      NAUTYEXTRA_TYPE_ULL v_without_dom_count = data[v].without_dom_count;
      v_without_dom += c_without_dom;
      int prospective = v_without + c_with;
      DEBUG(printf(" without_dom prospective %d [%Lu] cf %d [%Lu]\n",
                   prospective, c_with_count * v_without_count,
                   v_without_dom, v_without_dom_count));
      if (prospective >= v_without_dom) {
        MUL_NO_OVERFLOW_ULL(v_without_dom_count,
                            v_without_dom_count, c_without_dom_count);
        DEBUG(printf("  existing mul %Lu to [%Lu]\n",
                     c_without_dom_count, v_without_dom_count));
      }
      if (prospective <= v_without_dom) {
        NAUTYEXTRA_TYPE_ULL prospective_count;
        MUL_NO_OVERFLOW_ULL(prospective_count,
                            c_with_count, v_without_count);
        if (prospective < v_without_dom) {
          v_without_dom       = prospective;
          v_without_dom_count = prospective_count;
          DEBUG(printf("  new low parent without_dom %d [%Lu]\n",
                       v_without_dom, v_without_dom_count));
        } else {
          ADD_NO_OVERFLOW_ULL(v_without_dom_count,
                              v_without_dom_count, prospective_count);
          DEBUG(printf("  add prospective count [%Lu] to now [%Lu]\n",
                       prospective_count, v_without_dom_count));
        }
      }
      data[v].without_dom       = v_without_dom;
      data[v].without_dom_count = v_without_dom_count;

      /* v_without       = sum c_dom
         v_without_count = prod c_dom_count */
      data[v].without = v_without + c_dom;
      MUL_NO_OVERFLOW_ULL(data[v].without_count,
                          data[v].without_count, c_dom_count);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf(" stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  DEBUG(printf(" return %d count %Lu\n",
               total, (unsigned long long) total_count));
  info->domnum = total;
  info->count = total_count;
}


/* ------------------------------------------------------------------------- */
/* Independent Domination (Maximal Independent Sets)

   An independent dominating set of vertices is both independent (no
   adjacent vertices) and dominating (all graph vertices in the set or
   adjacent to one or more of the set).

   A set is independent dominating if and only if it is a maximal
   independent set.  A non-maximal independent set is not dominating since
   there is a vertex with no neighbours.  A maximal independent set is
   dominating since there are no neighbourless vertices which could be added.

   The independent dominating number is the minimum size of any independent
   dominating set.  This is also called the lower independence number, since
   it is the size of the smallest maximal independent set.

      domnum <= indomnum      <= indnum  <= updomnum
                = lowerindnum

   indom set iff maximal indset
   so indomnum smallest maximal indset
   minimal domset subset of minimal domset
   updomnum = maximum size of minimal domset
*/

/* ss is a set of vertices 0 to sg->nv-1.
   Return TRUE if ss is an independent dominating set (maximal independent
   set) in sg.  Return FALSE if not.

   This is like nautyextra_is_indset_sg() but additionaly if there is
   any vertex which could be added to ss and still be independent then ss is
   not maximal.  Such a vertex is some v not in ss and none of its
   neighbours in ss, hence undominated.
*/
int
nautyextra_is_indomset_sg (const sparsegraph *sg, const set *ss)
{
  int n = sg->nv;
  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;

  /* if from     in ss then want to check if it makes not independent
     if from not in ss then want to check if it could be, so not maximal
  */
  int from;
  for (from = 0; from < n; from++) {
    const int *from_eptr = ee + vv[from];
    int from_in_ss = ISELEMENT(ss,from);
    int from_degree = dd[from];
    int any_to = 0;
    int i;
    for (i=0; i<from_degree; i++) {
      int to = *from_eptr++;
      int to_in_ss = ISELEMENT(ss,to);
      if (from_in_ss && to_in_ss)
        return FALSE;  /* from,to both in ss, not independent */
      any_to |= to_in_ss;
    }

    /* from not in ss, and none of its neighbours in ss, so not dominated,
       or equivalently it could be added to the set and still be
       independent, so not maximal */
    if (!from_in_ss && !any_to)
      return FALSE;
  }
  return TRUE;
}

/* ENHANCE-ME: Prune the search like indsets_count. */
NAUTYEXTRA_TYPE_ULL
nautyextra_indomsets_count_sg (const sparsegraph *sg)
{
  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) return 1;  /*  empty set, one way */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_indomsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n, count += nautyextra_is_indomset_sg(sg,&ss));
  return count;
}

/* Return the number of independent dominating sets (maximal independent
   sets) in tree vpar[1..n].
   vpar must be a single tree, not a forest.
   This can be large, up to about 2^ceil(n/2) per Wilf.
   Currently overflow results in an abort().
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_indomsets_count_vpar (const int *vpar, int n)
{
  /* The code follows

     .    Herbert S. Wilf, "The Number of Maximal Independent Sets in a Tree",
     .    SIAM Journal on Algebraic and Discrete Methods, volume 7, number 1,
     .    January 1986, pages 125-130.
     .    https://www.math.upenn.edu/~wilf/website/Maximal%20independent%20sets%20in%20a%20tree.pdf

     Wilf calculates the number of maximal sets in a sub-tree at vertex v
     and below, with a count also of those sets which do not include v
     itself.  The counts then follow from corresponding counts at the
     children C and grandchilden G vertices.

     .    v without = prod(c any) - prod(c without)       without v
     .    v any     = v without + prod(g any)             with or without v

     The code maintains prod(c any), prod(c without), prod(g any) at a vertex.
     At a leaf the products in the parent are updated and that leaf discarded.

     When vertex v is not in the set there must be at least one child c, or
     the set is not maximal.  Hence excluding prod(c without) which is count
     combinations of child sets all of which don't have their respective c.

     When c is a leaf there are no maximal sets without that c (the sole
     maximal being c itself).  So prod(c without) is initialized to 0 at
     those v immediately above a leaf.

     Per Wilf, the result is linear time, making 4 passes over vpar, plus
     whatever the multiplications take.

     ---------------------------------------

     .        *
     .      /   \
     .    *       *
     .   / \     / \
     .  *   *   *   *

     3-->1<--2   sets 1 23

     .        1        path-6  count 5
     .       / \       1,3,5        with 2    = 1 way
     .      2   5      1,3,6        without 2 = 1 way
     .     /     \     1,4,6
     .    3       6    2,4,6        without 4 = 3
     .   /             2,5          with 4    = 2
     .  4

     .                            1       2*2 - 1*1 = 3   without
     .    without = 2-1 = 1      / \      3 + 2*1 = 5     any
     .    with    = 1+1 = 2     2   5        2,1
     .                         /     \
     .  1-0=1        2,1      3       6      1,0
     .  1+1=2                /         \
     .    0-0=0      1,0    4           *    1,0
     .    0+1=1            /
     .               1,0  *

     prod_c_any, prod_c_without, prod_g_any
     v_without = prod_c_with - prod_c_without
     v_any     = v_without + prod_g_any
     parent prod_c_any     *= v_any
     parent prod_c_without *= v_without
     parent prod_g_any     *= prod_c_any

     v_with                = prod c_without_notexcluded
     v_without             = prod c_any
     v_without_notexcluded = prod c_without_excluded
     v_without_excluded    = prod c_any - prod c_without
     (at least one child)
     final any = with + without
  */

  if (n == 0) return 1;
  if (n == 1) return 1;

  int i;
  struct data {
    int                  num_children;
    NAUTYEXTRA_TYPE_ULL  prod_c_any;
    NAUTYEXTRA_TYPE_ULL  prod_c_without;
    NAUTYEXTRA_TYPE_ULL  prod_g_any;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indomsets_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indomsets_count_vpar()");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].prod_c_any     = 1;
    data[i].prod_c_without = 1;
    data[i].prod_g_any     = 1;
  }
  for (i = 1; i <= n; i++) {
    if (data[i].num_children == 0) {
      data[vpar[i]].prod_c_without = 0;
    }
  }

  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d prods %Lu, %Lu, %Lu\n",
                 v, data[v].num_children,
                 (unsigned long long) data[v].prod_c_any,
                 (unsigned long long) data[v].prod_c_without,
                 (unsigned long long) data[v].prod_g_any));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      NAUTYEXTRA_TYPE_ULL prod_c_any     = data[v].prod_c_any;
      NAUTYEXTRA_TYPE_ULL prod_c_without = data[v].prod_c_without;
      NAUTYEXTRA_TYPE_ULL prod_g_any     = data[v].prod_g_any;
      assert(prod_c_any >= prod_c_without);
      NAUTYEXTRA_TYPE_ULL v_without = prod_c_any - prod_c_without;
      NAUTYEXTRA_TYPE_ULL v_any;
      ADD_NO_OVERFLOW_ULL(v_any, v_without, prod_g_any);

      DEBUG(printf("consider v=%d prods %Lu, %Lu, %Lu  = %Lu without, %Lu any\n",
                   v,
                   (unsigned long long) prod_c_any,
                   (unsigned long long) prod_c_without,
                   (unsigned long long) prod_g_any,
                   (unsigned long long) v_without,
                   (unsigned long long) v_any));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  root\n"));
        MUL_NO_OVERFLOW_ULL(total_count, total_count, v_any);
        break;
      }

      /* parent prod_c_any *= v_any */
      MUL_NO_OVERFLOW_ULL(data[v].prod_c_any,
                          data[v].prod_c_any, v_any);

      /* parent prod_c_without *= v_without */
      MUL_NO_OVERFLOW_ULL(data[v].prod_c_without,
                          data[v].prod_c_without, v_without);

      /* parent prod_g_any *= prod_c_any */
      MUL_NO_OVERFLOW_ULL(data[v].prod_g_any,
                          data[v].prod_g_any, prod_c_any);

      DEBUG(printf("  to parent v=%d giving %Lu, %Lu, %Lu\n",
                   v,
                   (unsigned long long) data[v].prod_c_any,
                   (unsigned long long) data[v].prod_c_without,
                   (unsigned long long) data[v].prod_g_any));

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d  count %Lu, %Lu, %Lu\n",
                     v,
                     (unsigned long long) data[v].prod_c_any,
                     (unsigned long long) data[v].prod_c_without,
                     (unsigned long long) data[v].prod_g_any));
        break;
      }
    }
  }

  return total_count;
}

/* Return the largest number of independent dominating sets (maximal
   independent sets) which occurs among trees of n vertices, per Wilf
   reference above.
   This is the maximum value which nautyextra_indomsets_count_vpar()
   might return for a tree of n vertices.
   n=0 is reckoned as having 1 maximal set (the empty set).

   Wilf gives stars with 2-vertex arms as example trees attaining this maximum.
   See nautyextra_make_star_arms2_vpar() to make those.


   GP-DEFINE  f(n) = if(n==0,1, n%2==0,2^(n/2-1)+1, 2^((n-1)/2));
   GP-Test  vector(10,n,n--; f(n)) ==                            \
   GP-Test  vector(10,n,n--; if(n==0,1, 2^ceil(n/2-1)+1-(n%2)))
   GP-Test  vector(10,n,n--; f(n)) == \
   GP-Test  vector(10,n,n--; if(n==0,1, 2^floor((n-1)/2) + if(n%2==0,1)))

   forest same when n odd, for even bigger so 2^floor(n/2) all n
   GP-DEFINE  forest_f(n) = {
   GP-DEFINE    my(m=0,mv=[]);
   GP-DEFINE    forpart(v=n,
   GP-DEFINE       my(new_m=prod(i=1,#v,f(v[i])));
   GP-DEFINE       if(new_m>m, m=new_m;mv=v));
   GP-DEFINE    [m, Vec(mv)];
   GP-DEFINE  }
   GP-Test  f(2) == 2
   GP-Test  f(18) == 257
   GP-Test  f(19) == 512
   for(n=0,20,print("n="n"  "f(n)"  "forest_f(n)))
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_indomsets_count_tree_max (int n)
{
  if (n==0) return 1;
  NAUTYEXTRA_TYPE_ULL count = 1ULL << ((n-1)>>1);
  /* FIXME: some CPUs wrap around rather than go to 0 */
  if (count == 0) INTERNAL_overflow();
  return count + (1-(n&1));
}

void
nautyextra_indomnum_and_count_sg (const sparsegraph *sg, struct nautyextra_indomnum_and_count *info)
{
  DEBUG(printf("nautyextra_indomnum_and_count_sg() n=%d\n", sg->nv));
  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    info->indomnum = 0; /*  empty set, one way */
    info->count = 1;
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_indomnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int indomnum = n+1;
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     if (nautyextra_is_indomset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < indomnum) {
         DEBUG(printf(" new high %s\n", nautyextra_set_str(&ss,n)));
         indomnum = ss_size;
         count = 0;
       }
       if (ss_size == indomnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end indomnum %d count %u\n", indomnum, count));
  if (! count) {
    /* nothing at all inal dominates, due to isolated vertex */
    assert (indomnum == n+1);
    indomnum = -1;
  }
  info->indomnum = indomnum;
  info->count = count;
}


/* Return the independent domination number of vpar[1..n].
   This is the number of vertices in the smallest independent dominating set
   (so smallest maximal independent set).

   vpar can be a forest, in which case the effect is sum of independent
   domination numbers of its component trees.
*/
int
nautyextra_indomnum_vpar (const int *vpar, int n)
{
  int i;
  struct data {
    int  num_children;
    int  with;
    int  without_dom;
    int  without_dom_any;
    int  without_undom;
    int  without_undom_any;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indomnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indomnum_vpar() ");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with              = 1;
    data[i].without_undom_any = 1;
  }

  int total_indomnum = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      int  c_with              = data[v].with;
      int  c_without_dom       = data[v].without_dom;
      int  c_without_dom_any   = data[v].without_dom_any;
      int  c_without_undom     = data[v].without_undom;
      int  c_without_undom_any = data[v].without_undom_any;

      DEBUG(printf("consider v=%d child\n", v));
      DEBUG(printf("  with %d without_dom %d [%d] without_undom %d [%d]\n",
                   c_with,
                   c_without_dom,c_without_dom_any,
                   c_without_undom,c_without_undom_any));
      assert (data[v].num_children == 0);

      /* c_dom = min(c_with, c_without_dom)
         both counts when equal min
         each c can be with or without, but must be dominated from below
      */
      int c_dom = c_with;
      if (c_without_dom_any && c_without_dom < c_dom)
        c_dom = c_without_dom;
      DEBUG(printf("  c_dom %d\n", c_dom));

      data[v].num_children = -1;  /* drop this leaf */
      v = vpar[v];                /* to parent */
      if (! v) {
        DEBUG(printf("final c_dom %d\n", c_dom));
        total_indomnum += c_dom;
        assert (0 <= total_indomnum && total_indomnum <= n);
        break;
      }

      /* before update */
      assert (data[v].without_dom_any || data[v].without_undom_any);
      int v_without_dom     = data[v].without_dom;
      int v_without_dom_any = data[v].without_dom_any;
      int v_without_undom     = data[v].without_undom;
      int v_without_undom_any = data[v].without_undom_any;

      /* v_without = min(v_without_dom, v_without_undom) */
      int v_without = v_without_undom;
      if (!v_without_undom_any
          || (v_without_dom_any && v_without_dom < v_without))
        v_without = v_without_dom;
      DEBUG(printf("  v_without_dom %d [%d] v_without_undom %d [%d]  is v_without %d\n",
                   v_without_dom, v_without_dom_any,
                   v_without_undom, v_without_undom_any,
                   v_without));

      {
        /* c_without = min(c_without_dom, c_without_undom) */
        assert (c_without_dom_any || c_without_undom_any);
        int c_without = c_without_undom;
        if (!c_without_undom_any
            || (c_without_dom_any && c_without_dom < c_without))
          c_without = c_without_dom;
        DEBUG(printf("  c_without %d\n", c_without));

        /* v_with       = 1 + sum c_without
           v present means each c can be dom or undom, but must without to be
           independent */
        data[v].with += c_without;
        DEBUG(printf("  set v_with          %d\n", data[v].with));
      }

      /* v without_undom = sum(c_without_dom) */
      data[v].without_undom     = v_without_undom + c_without_dom;
      data[v].without_undom_any = v_without_undom_any & c_without_dom_any;
      DEBUG(printf("  set v_without_undom %d [%d]\n",
                   data[v].without_undom, data[v].without_undom_any));

      /* v_without_dom   = min (>= one c_with, others c_dom)
         prospective = c_with + (others c_dom = v_without before this)
         vs
         v_without_dom += c_without_dom    so not include prospective c_with
      */
      {
        int prospective = c_with + v_without;
        v_without_dom     += c_without_dom;
        v_without_dom_any &= c_without_dom_any;
        DEBUG(printf("  prospective %d  vs %d [%d]\n",
                     prospective, v_without_dom, v_without_dom_any));

        /* always take the prospective when no sets v_without_dom_any */
        if (prospective <= v_without_dom || !v_without_dom_any) {
          if (prospective < v_without_dom || !v_without_dom_any) {
            v_without_dom = prospective;
          }
          v_without_dom_any = 1;
        }
        data[v].without_dom     = v_without_dom;
        data[v].without_dom_any = v_without_dom_any;
        DEBUG(printf("  set v_without_dom   %d [%d]\n",
                     v_without_dom, v_without_dom_any));
      }

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  return total_indomnum;
}


/* Set info->indomnum to the independent domination number of vpar[1..n] and
   set info->count to the number of sets of that size.

   vpar can be a forest, in which the effect is info->indomnum sum of
   independent domination numbers and info->count product of number of sets
   of the component trees.
*/
void
nautyextra_indomnum_and_count_vpar (const int *vpar, int n,
                                    struct nautyextra_indomnum_and_count *info)
{
  /* v_with          = 1 + sum c_without
     v_without_undom =     sum c_without_dom
     v_without       =     sum c_dom
     v_without_dom   = min (>= one c_with, others c_without_dom)

     v_without = min(v_without_dom,v_without_undom)
     c_dom = min(c_with, c_without_dom)

     with            \ dom              \
     without_dom     /      \ without   | any
     without_undom          /           /

     "without" in all cases has the children all dominated by something
     further below.  When a child is a leaf there is nothing below it which
     can dominate so its parent has without_undom_count==0 as there are no
     sets with the parent undominated but its children dominated.

     without dom - undom unbounded +ve or -ve maybe ...
     with - without unbounded +ve or -ve maybe ...
  */

  int i;
  struct data {
    int                  num_children;
    int                  with;
    int                  without_dom;
    int                  without_undom;
    NAUTYEXTRA_TYPE_ULL  with_count;
    NAUTYEXTRA_TYPE_ULL  without_dom_count;
    NAUTYEXTRA_TYPE_ULL  without_undom_count;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_indomnum_and_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indomnum_and_count_vpar() ");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with                = 1;
    data[i].with_count          = 1;
    data[i].without_undom_count = 1;
  }

  int total_indomnum = 0;
  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      int  c_with          = data[v].with;
      int  c_without_dom   = data[v].without_dom;
      int  c_without_undom = data[v].without_undom;
      NAUTYEXTRA_TYPE_ULL  c_with_count          = data[v].with_count;
      NAUTYEXTRA_TYPE_ULL  c_without_dom_count   = data[v].without_dom_count;
      NAUTYEXTRA_TYPE_ULL  c_without_undom_count = data[v].without_undom_count;

      DEBUG(printf("consider v=%d child\n", v));
      DEBUG(printf("  with %d [%Lu] without_dom %d [%Lu] without_undom %d [%Lu]\n",
                   c_with, c_with_count,
                   c_without_dom, c_without_dom_count,
                   c_without_undom, c_without_undom_count));
      assert (data[v].num_children == 0);


      /* c_dom = min(c_with, c_without_dom)
         both counts when equal min
         each c can be with or without
         can have c_without_dom_count==0, eg. at leaf nothing below to
         dominate it
      */
      int                  c_dom       = c_with;
      NAUTYEXTRA_TYPE_ULL  c_dom_count = c_with_count;
      if (c_without_dom_count && c_without_dom < c_dom) {
        c_dom       = c_without_dom;
        c_dom_count = c_without_dom_count;
      } else if (c_without_dom == c_dom) {
        ADD_NO_OVERFLOW_ULL(c_dom_count,
                            c_dom_count, c_without_dom_count);
      }
      DEBUG(printf("  c_dom      %d [%Lu]\n", c_dom, c_dom_count));


      data[v].num_children = -1;  /* drop this leaf */
      v = vpar[v];                /* to parent */
      if (! v) {
        total_indomnum += c_dom;
        MUL_NO_OVERFLOW_ULL(total_count, total_count, c_dom_count);
        assert (total_count != 0);
        assert (0 <= total_indomnum && total_indomnum <= n);
        break;
      }

      {
        /* c_without = min(c_without_dom, c_without_undom)
           can have c_without_undom_count==0 for no sets of without_undom,
           eg. immediate parent of a leaf (the leaf must be present or that
           leaf would be undominated)
        */
        int                  c_without       = c_without_dom;
        NAUTYEXTRA_TYPE_ULL  c_without_count = c_without_dom_count;
        if (c_without_undom_count && c_without_undom < c_without) {
          c_without       = c_without_undom;
          c_without_count = c_without_undom_count;
        } else if (c_without_undom == c_without) {
          ADD_NO_OVERFLOW_ULL(c_without_count,
                              c_without_count, c_without_undom_count);
        }
        DEBUG(printf("  c_without  %d [%Lu]\n", c_without, c_without_count));

        /* v_with       = 1 + sum c_without
           v present means each c can be dom or undom, but must without to be
           independent */
        data[v].with += c_without;
        MUL_NO_OVERFLOW_ULL(data[v].with_count,
                            data[v].with_count, c_without_count);
      }

      /* v without_undom = sum(c_without_dom)
         and leave v_without_undom as value before update
      */
      int                 v_without_undom = data[v].without_undom;
      NAUTYEXTRA_TYPE_ULL v_without_undom_count = data[v].without_undom_count;
      data[v].without_undom = v_without_undom + c_without_dom;
      MUL_NO_OVERFLOW_ULL(data[v].without_undom_count,
                          v_without_undom_count, c_without_dom_count);
      DEBUG(printf("  v without_undom to %d [%Lu]  from child c_without_dom\n",
                   data[v].without_undom, data[v].without_undom_count));

      /* v_without = min(v_without_dom, v_without_undom) calculated
         before update, used for v_without = sum c_dom
      */
      int                  v_without_dom       = data[v].without_dom;
      NAUTYEXTRA_TYPE_ULL  v_without_dom_count = data[v].without_dom_count;
      int                  v_without       = v_without_dom;
      NAUTYEXTRA_TYPE_ULL  v_without_count = v_without_dom_count;
      if (v_without_undom_count && v_without_undom < v_without) {
        v_without       = v_without_undom;
        v_without_count = v_without_undom_count;
      } else if (v_without_undom == v_without) {
        ADD_NO_OVERFLOW_ULL(v_without_count,
                            v_without_count, v_without_undom_count);
      }
      DEBUG(printf("  v_without   %d [%Lu]\n", v_without, v_without_count));

      /* v_without_dom   = min (>= one c_with, others c_dom)
         prospective = c_with + (others c_dom = v_without before this)
         vs
         v_without_dom += c_without_dom    so not include prospective c_with
      */
      {
        v_without_dom += c_without_dom;
        MUL_NO_OVERFLOW_ULL(v_without_dom_count,
                            v_without_dom_count, c_without_dom_count);

        int prospective = c_with + v_without;
        DEBUG(printf("  v without_dom prospective %d [%Lu] cf %d [%Lu]\n",
                     prospective, c_with_count * v_without_count,
                     v_without_dom, v_without_dom_count));

        /* always take the prospective when no sets v_without_dom */
        int v_without_dom_is_empty = (v_without_dom_count == 0);
        if (prospective <= v_without_dom || v_without_dom_is_empty) {
          NAUTYEXTRA_TYPE_ULL  prospective_count;
          MUL_NO_OVERFLOW_ULL(prospective_count, c_with_count, v_without_count);
          assert (prospective_count != 0);

          if (prospective < v_without_dom || v_without_dom_is_empty) {
            v_without_dom       = prospective;
            v_without_dom_count = prospective_count;
            DEBUG(printf("    new low parent without_dom %d [%Lu]\n",
                         v_without_dom, v_without_dom_count));
          } else {
            ADD_NO_OVERFLOW_ULL(v_without_dom_count,
                                v_without_dom_count, prospective_count);
            DEBUG(printf("    add prospective count [%Lu] to now [%Lu]\n",
                         prospective_count, v_without_dom_count));
          }
        }
        data[v].without_dom       = v_without_dom;
        data[v].without_dom_count = v_without_dom_count;
      }

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

  info->indomnum = total_indomnum;
  info->count     = total_count;
}


/* ------------------------------------------------------------------------- */
/* Total Domination */

/* ss is a set of vertices 0 to sg->nv-1.
   sg is an undirected sparsegraph without self-loops.
   Return TRUE if ss is a total dominating set in sg.
   Return FALSE if not.

   A set total dominates if every vertex has some neighbour in the set.
   This differs from ordindinary nautyextra_is_domset_sg() etc in that a
   vertex in ss does not dominate itself, it must have some other neighbour.

   Currently a vertex in ss with a self-loop is reckoned as dominated by
   itself.  Not sure if that's a good idea.  Don't rely on it yet.
*/
int
nautyextra_is_totdomset_sg (const sparsegraph *sg, const set *ss)
{
  /* cf setnbhd() getting all neighbours of all in a set */

  int n = sg->nv;
  DEBUG(printf("nautyextra_is_totdomset_sg() n=%d  %s\n",
               n, nautyextra_set_str(ss,n)));

  const int    *dd = sg->d;
  const int    *ee = sg->e;
  const size_t *vv = sg->v;
  int v;
  for (v=0; v<n; v++) {
    const int *eptr = ee + vv[v];
    int degree = dd[v];
    int i;
    for (i=0; i<degree; i++) {
      int to = *eptr++;
      if (ISELEMENT(ss,to)) goto NEXT_V;  /* neighbour "to" in ss, good */
    }
    return FALSE;  /* v is not dominated, no neighbour in ss */
  NEXT_V: ;
  }
  DEBUG(printf("  is totdomset\n"));
  return TRUE;
}

/* sg is an undirected sparsegraph.
   Return the number of total dominating sets in sg.

   Currently this is a naive search of all sets, so is suitable only for a
   few vertices, and currently only uses 1-word sets, so must have sg->nv <=
   WORDSIZE many vertices.
*/
NAUTYEXTRA_TYPE_ULL
nautyextra_totdomsets_count_sg (const sparsegraph *sg)
{
  setword count = 0;
  setword ss;
  int n = sg->nv;

  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_totdomsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_totdomset_sg (sg,&ss));
  return count;
}

/* sg is an undirected graph, possibly with self-loops.
   Set *info to the total domination number and count of sets of this size
   in sg.
   If sg contains an isolated vertex then there are no sets of any size
   which can total dominate.  In that case return totaldomnum = -1,
   count = 0.

   Currently this is a naive search of all sets, so is suitable only for a
   small number of vertices, and currently only uses 1-word sets, so must
   have sg->nv <= WORDSIZE many vertices.

   ENHANCE-ME: At any leaf vertex must take its attachment.  Unless a path-2
   the leaf can be excluded, since between dominating the attachment with
   the leaf or with another vertex the other vertex will be a set size <=
   taking the leaf.
*/
void
nautyextra_totdomnum_and_count_sg (const sparsegraph *sg,
                                   struct nautyextra_totdomnum_and_count *info)
{
  DEBUG(printf("nautyextra_totdomnum_and_count_sg() n=%d\n", sg->nv));
  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    info->totdomnum = 0; /*  empty set, one way */
    info->count = 1;
    return;
  }
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_totdomnum_and_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int totdomnum = n+1;
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     if (nautyextra_is_totdomset_sg(sg,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size < totdomnum) {
         DEBUG(printf(" new high %s\n", nautyextra_set_str(&ss,n)));
         totdomnum = ss_size;
         count = 0;
       }
       if (ss_size == totdomnum) {
         count++;
         DEBUG(printf(" count %u\n", count));
       }
     });
  DEBUG(printf(" end totdomnum %d count %u\n", totdomnum, count));
  if (! count) {
    /* nothing at all total dominates, due to isolated vertex */
    assert (totdomnum == n+1);
    totdomnum = -1;
  }
  info->totdomnum = totdomnum;
  info->count = count;
}

/* Return the total domination number of tree vpar[1..n].
   vpar can be a forest, in which case the result is sum of total domination
   numbers of its component trees.
   An isolated vertex cannot be total dominated.  If any such in vpar then
   return -1.
 */
int
nautyextra_totdomnum_vpar (const int *vpar, int n)
{
  DEBUG(printf("nautyextra_totdomnum_vpar() ");
        nautyextra_print_vpar(vpar,n));
  assert (n >= 0);

  const int DOMINATED   = 1;
  const int PRESENT     = 2;
  const int HAVE_CHILD  = 4;
  struct data {
    int  num_children;
    int  state;          /* mask 1=dominated, 2=present */
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_totdomnum_vpar");
#endif
  int totdomnum = 0;
  int i;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
  }

  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      int c_state     = data[v].state;
      int c_dominated = c_state & DOMINATED;
      DEBUG(printf("consider v=%d (parent %d)  dominated %d present %d have_child %d\n",
                   v, vpar[v],
                   c_dominated, c_state & PRESENT, c_state & HAVE_CHILD));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        totdomnum += 1 - c_dominated; /* if not dom then take child */
        DEBUG(printf("  root, to totdomnum %d\n", totdomnum));
        if (! (c_state & (DOMINATED | HAVE_CHILD))) {
          DEBUG(printf("  not dominated and no child, is isolated vertex\n"));
          return -1;
        }
        break;
      }

      int v_state = data[v].state;

      /* c already dominated or v already present don't taken parent,
         otherwise c undominated and v not yet present, increase totdomnum
      */
      totdomnum += 1 - (c_dominated | ((v_state & PRESENT) >> 1));

      /* parent dominated when child present,
         parent present if child undominated,
         parent has child
      */
      data[v].state = v_state | (((c_state & PRESENT) >> 1)
                                 | ((1-c_dominated) << 1)
                                 | HAVE_CHILD);

      DEBUG(printf("  to parent dominated %d present %d have_child %d  totdomnum %d\n",
                   data[v].state & DOMINATED,
                   data[v].state & PRESENT, data[v].state & HAVE_CHILD,
                   totdomnum));

      assert(data[v].num_children > 0);
      if (--data[v].num_children) break;
    }
  }

  DEBUG(printf("  final %d\n", totdomnum));
  return totdomnum;
}

/* Set info->totdomnum to the total domination number of tree vpar[1..n] and
   set info->count to the number of total dominating sets of that size.

   vpar can be a forest, in which case the result is sum total domination
   numbers and product count of sets of its component trees.

   An isolated vertex cannot be total dominated.  If any such in vpar then
   return info->totdomnum = -1 and info->count = 0.
 */
void
nautyextra_totdomnum_and_count_vpar (const int *vpar, int n,
                                     struct nautyextra_totdomnum_and_count *info)
{
  int i;
  struct data {
    int                  num_children;
    int                  with;
    int                  with_dom;
    int                  with_undom;
    int                  without;
    int                  without_dom;
    int                  without_undom;
    NAUTYEXTRA_TYPE_ULL  with_count;
    NAUTYEXTRA_TYPE_ULL  with_dom_count;
    NAUTYEXTRA_TYPE_ULL  with_undom_count;
    NAUTYEXTRA_TYPE_ULL  without_count;
    NAUTYEXTRA_TYPE_ULL  without_dom_count;
    NAUTYEXTRA_TYPE_ULL  without_undom_count;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert (n <= MAXN);
#else
  WORKSPACE(struct data,data,n+1,"nautyextra_totdomnum_and_count_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domsets_count_vpar() ");
        nautyextra_print_vpar(vpar,n));

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].with                = 1;
    data[i].with_count          = 1;
    data[i].with_dom            = 1;  /* with_dom_count = 0 */
    data[i].with_undom          = 1;
    data[i].with_undom_count    = 1;

    data[i].without_count       = 1;
    data[i].without_undom_count = 1;  /* without_dom_count = 0 */
  }

  int total_totdomnum = 0;
  NAUTYEXTRA_TYPE_ULL total_count = 1;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d\n", v, data[v].num_children));
    if (data[v].num_children != 0) continue;

    for (;;) {
      int  c_with          = data[v].with;
      int  c_with_dom      = data[v].with_dom;
      int  c_with_undom    = data[v].with_undom;
      int  c_without       = data[v].without;
      int  c_without_dom   = data[v].without_dom;
      int  c_without_undom = data[v].without_undom;

      NAUTYEXTRA_TYPE_ULL  c_with_count          = data[v].with_count;
      NAUTYEXTRA_TYPE_ULL  c_with_dom_count      = data[v].with_dom_count;
      NAUTYEXTRA_TYPE_ULL  c_with_undom_count    = data[v].with_undom_count;
      NAUTYEXTRA_TYPE_ULL  c_without_count       = data[v].without_count;
      NAUTYEXTRA_TYPE_ULL  c_without_dom_count   = data[v].without_dom_count;
      NAUTYEXTRA_TYPE_ULL  c_without_undom_count = data[v].without_undom_count;

      DEBUG(printf("consider v=%d child with %d [%Lu] with_dom %d [%Lu] with_undom %d [%Lu]\n",
                   v,
                   c_with,c_with_count,
                   c_with_dom,c_with_dom_count,
                   c_with_undom,c_with_undom_count));
      DEBUG(printf("           without %d [%Lu] without_dom %d [%Lu] without_undom %d [%Lu]\n",
                   c_without,c_without_count,
                   c_without_dom,c_without_dom_count,
                   c_without_undom,c_without_undom_count));
      assert (data[v].num_children == 0);
      if (c_with_count && c_with_dom_count && c_with_undom_count)
        assert(c_with == MIN(c_with_dom,c_with_undom));
      if (c_without_count && c_without_dom_count && c_without_undom_count)
        assert(c_without == MIN(c_without_dom,c_without_undom));
      /* assert(c_with_undom_count > 0); */
      /* assert(c_without_undom_count > 0); */


      /* c_dom = min(c_with_dom, c_without_dom)
         both counts when equal min
         each c can be with or without, but must be dominated from below
      */
      int                  c_dom       = c_with_dom;
      NAUTYEXTRA_TYPE_ULL  c_dom_count = c_with_dom_count;
      if (c_without_dom_count && c_without_dom < c_dom) {
        c_dom       = c_without_dom;
        c_dom_count = c_without_dom_count;
      } else if (c_without_dom == c_dom) {
        ADD_NO_OVERFLOW_ULL(c_dom_count,
                            c_dom_count, c_without_dom_count);
      }
      DEBUG(printf("  c_dom   %d [%Lu]\n", c_dom, c_dom_count));

      data[v].num_children = -1;  /* drop this leaf */
      v = vpar[v];                /* to parent */
      if (! v) {
        if (c_dom_count == 0) {
          DEBUG(printf("  root c_dom_count = %Lu, no totdomsets\n", c_dom_count));
          total_totdomnum = -1;
          total_count = 0;
          goto END;
        }
        total_totdomnum += c_dom;
        total_count     *= c_dom_count;
        DEBUG(printf("  root accumulate to %d [%Lu]\n", total_totdomnum, total_count));
        break;
      }

      /* c_undom = min(c_with_undom, c_without_undom) */
      int                  c_undom       = c_with_undom;
      NAUTYEXTRA_TYPE_ULL  c_undom_count = c_with_undom_count;
      if (c_without_undom_count && c_without_undom < c_undom) {
        c_undom       = c_without_undom;
        c_undom_count = c_without_undom_count;
      } else if (c_without_undom == c_undom) {
        ADD_NO_OVERFLOW_ULL(c_undom_count,
                            c_undom_count, c_without_undom_count);
      }
      DEBUG(printf("  c_undom %d [%Lu]\n", c_undom, c_undom_count));

      /* c_any = min(c_dom, c_undom) */
      int                  c_any       = c_dom;
      NAUTYEXTRA_TYPE_ULL  c_any_count = c_dom_count;
      if (c_undom_count && c_undom < c_any) {
        c_any       = c_undom;
        c_any_count = c_undom_count;
      } else if (c_undom == c_any) {
        ADD_NO_OVERFLOW_ULL(c_any_count,
                            c_any_count, c_undom_count);
      }
      DEBUG(printf("  c_any   %d [%Lu]\n", c_any, c_any_count));


      int                 v_with       = data[v].with;
      NAUTYEXTRA_TYPE_ULL v_with_count = data[v].with_count;
      int                 v_without       = data[v].without;
      NAUTYEXTRA_TYPE_ULL v_without_count = data[v].without_count;

      /* v_with       = 1 + sum c_any
         v present means each c can be dom or undom */
      data[v].with += c_any;
      MUL_NO_OVERFLOW_ULL(data[v].with_count,
                          data[v].with_count, c_any_count);
      DEBUG(printf("  v with was %d [%Lu] now %d [%Lu]\n",
                   v_with, v_with_count,
                   data[v].with, data[v].with_count));

      /* v without = sum(c_dom)
         v absent means all c must be dom from below */
      data[v].without += c_dom;
      MUL_NO_OVERFLOW_ULL(data[v].without_count,
                          data[v].without_count, c_dom_count);

      /* v with_undom = 1 + sum(c_without) */
      data[v].with_undom += c_without;
      MUL_NO_OVERFLOW_ULL(data[v].with_undom_count,
                          data[v].with_undom_count, c_without_count);
      DEBUG(printf("  v with_undom to %d [%Lu]\n",
                   data[v].with_undom, data[v].with_undom_count));

      /* v without_undom = sum(c_without_dom)
         v absent means all c must be dominated from below */
      data[v].without_undom += c_without_dom;
      MUL_NO_OVERFLOW_ULL(data[v].without_undom_count,
                          data[v].without_undom_count, c_without_dom_count);
      DEBUG(printf("  v without_undom to %d [%Lu]  from child c_without_dom\n",
                   data[v].without_undom, data[v].without_undom_count));


      /* v_without_dom   = min (>= one c_with_dom, others c_dom)
         prospective = c_with_dom + (others c_dom = v_without before this child)
         vs
         v_without_dom += c_without_dom    so not include prospective c_with_dom
      */
      {
        int                 v_without_dom       = data[v].without_dom;
        NAUTYEXTRA_TYPE_ULL v_without_dom_count = data[v].without_dom_count;
        v_without_dom += c_without_dom;
        int prospective = c_with_dom + v_without;

        DEBUG(printf("  v without_dom prospective %d [%Lu] cf %d [%Lu]\n",
                     prospective, c_with_dom_count * v_without_count,
                     v_without_dom, v_without_dom_count));

        DEBUG(printf("    existing %Lu * c_without_dom_count %Lu to [%Lu]\n",
                     v_without_dom_count, c_without_dom_count,
                     v_without_dom_count *c_without_dom_count));
        MUL_NO_OVERFLOW_ULL(v_without_dom_count,
                            v_without_dom_count, c_without_dom_count);

        NAUTYEXTRA_TYPE_ULL prospective_count;
        MUL_NO_OVERFLOW_ULL(prospective_count,
                            c_with_dom_count, v_without_count);
        if (prospective_count) {
          if (prospective >= v_without_dom) {
          }
          if (prospective <= v_without_dom || !v_without_dom_count) {
            if (prospective < v_without_dom || !v_without_dom_count) {
              v_without_dom       = prospective;
              v_without_dom_count = prospective_count;
              DEBUG(printf("    new low parent without_dom %d [%Lu]\n",
                           v_without_dom, v_without_dom_count));
            } else {
              ADD_NO_OVERFLOW_ULL(v_without_dom_count,
                                  v_without_dom_count, prospective_count);
              DEBUG(printf("    add prospective count [%Lu] to now [%Lu]\n",
                           prospective_count, v_without_dom_count));
            }
          }
        }
        data[v].without_dom       = v_without_dom;
        data[v].without_dom_count = v_without_dom_count;
      }


      /* v_with_dom   = 1 + min (>= one c_with, others c_any)
         prospective = 1 + c_with + others c_any
         .                      1 + others c_any = v_with before this child
         vs
         v_with_dom += c_without     so not include prospective c_with
      */
      {
        int                 v_with_dom       = data[v].with_dom;
        NAUTYEXTRA_TYPE_ULL v_with_dom_count = data[v].with_dom_count;
        v_with_dom += c_without;
        int prospective = c_with + v_with;
        DEBUG(printf("  v with_dom prospective %d+%d=%d [%Lu] cf %d [%Lu]\n",
                     c_with,v_with,prospective, c_without_count * v_with_count,
                     v_with_dom, v_with_dom_count));

        MUL_NO_OVERFLOW_ULL(v_with_dom_count,
                            v_with_dom_count, c_without_count);
        DEBUG(printf("    existing mul %Lu to [%Lu]\n",
                     c_without_count, v_with_dom_count));

        NAUTYEXTRA_TYPE_ULL prospective_count;
        MUL_NO_OVERFLOW_ULL(prospective_count,
                            c_with_count, v_with_count);

        if (prospective_count) {
          if (prospective >= v_with_dom) {
          }
          if (prospective <= v_with_dom || !v_with_dom_count) {
            if (prospective < v_with_dom || !v_with_dom_count) {
              v_with_dom       = prospective;
              v_with_dom_count = prospective_count;
              DEBUG(printf("    new low parent with_dom %d [%Lu]\n",
                           v_with_dom, v_with_dom_count));
            } else {
              ADD_NO_OVERFLOW_ULL(v_with_dom_count,
                                  v_with_dom_count, prospective_count);
              DEBUG(printf("    add prospective count [%Lu] to now [%Lu]\n",
                           prospective_count, v_with_dom_count));
            }
          }
        }
        data[v].with_dom       = v_with_dom;
        data[v].with_dom_count = v_with_dom_count;
      }


      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d\n", v));
        break;
      }
    }
  }

 END:
  info->totdomnum = total_totdomnum;
  info->count     = total_count;


  /* path 6 totdomnum=4 in count 4 ways
     .     v
     1--2--3--4--5--6
     @  @  .  .  @  @
     @  @  .  @  @  .
     .  @  @  .  @  @
     .  @  @  @  @  .

     path 3 totdomnum=2 in count 2 ways
     .     v
     1--2--3
     @  @  .
     .  @  @
  */
}


/* ------------------------------------------------------------------------- */
/* Relabelling */

/* Change vpar[1..n] to swap vertex labels u,v.
 */
void
nautyextra_relabel_swap_vpar (int *vpar, int n, int u, int v)
{
  assert (n >= 0);
  assert (1 <= u && u <= n);
  assert (1 <= v && v <= n);
  int i;
  for (i = 1; i <= n; i++) {
    int p = vpar[i];
    if      (p==u) vpar[i]=v;
    else if (p==v) vpar[i]=u;
  }
  int t = vpar[u]; vpar[u] = vpar[v]; vpar[v] = t;
}

/* Copy vpar_in[1..n] to vpar_out[1..n] with relabelling according to
   permutation perm[1..n].
   perm must be a permutation of the integers 1..n with perm[0]==0.
   Each vertex v in vpar_in becomes vertex perm[v] in vpar_out.
   Must have vpar_in different location than vpar_out (no in-place relabel).

   Cf nauty relabel() and relabel_sg() which are in-place (via workspace)
   and with lab[] labelling permuted too.
*/
void
nautyextra_relabel_perm_vpar (const int *vpar_in, int n, const int *perm, int *vpar_out)
{
  DEBUG(printf("nautyextra_relabel_perm_vpar() n=%d  %p -> %p\n",
               n, vpar_in, vpar_out));
  assert (n >= 0);
  assert (vpar_in != vpar_out);
  assert (perm[0] == 0);

  int i;
  if (n) {
    assert (perm[0] == 0);
    for (i = 1; i <= n; i++) {
      assert (0 <= vpar_in[i] && vpar_in[i] <= n);
      assert (1 <= perm[i] && perm[i] <= n);
      assert (0 <= perm[vpar_in[i]] && perm[vpar_in[i]] <= n);
      vpar_out[perm[i]] = perm[vpar_in[i]];
    }
  }
}

/* Set perm[1..n] to a random permutation of the integers 1..n, and perm[0]=0.
   Cf nauty ranperm() which is for integers 0..n-1.
 */
void
nautyextra_ranperm_vpar (int *perm, int n)
{
  int i;
  for (i = 0; i <= n; i++)
    perm[i] = i;

  for (i = n; i >= 2; i--) {
    int to; GETKRAN(i,to); to++;   /* 1 to i inclusive */
    /* printf ("i=%d to %d\n", i, to); */

    assert(1 <= i  && i <= n);
    assert(1 <= to && to <= n);
    int temp = perm[i];
    perm[i] = perm[to];
    perm[to] = temp;
  }

#if WANT_DEBUG
  printf("ranperm [%d]", perm[0]);
  for (i = 1; i <= n; i++) {
    printf(" %d", perm[i]);
  }
  putchar('\n');
#endif
}

/* Copy tree vpar_in[1..n] to vpar_out[1..n] with random vertex relabelling.
   Must have vpar_in different location than vpar_out (no in-place relabel).

   Cf nauty ranlabg.c static ranrelabel() for graph.
*/
void
nautyextra_ranrelabel_vpar (const int *vpar_in, int n, int *vpar_out)
{
#if MAXN
  int perm[MAXN+1];    /* 1 to n inclusive */
#else
  WORKSPACE(int,perm,n+1,"nautyextra_ranrelabel_vpar");
#endif
  nautyextra_ranperm_vpar(perm,n);
  nautyextra_relabel_perm_vpar(vpar_in,n, perm, vpar_out);
}


/* ------------------------------------------------------------------------- */
/* Rerooting */

/* Change vpar[1..n] so that v is the root of the tree, or root of its
   component tree if vpar is a forest.

   The neighbours (children plus parent) at all vertices are unchanged, just
   some switched from parent to child or vice versa.
*/
void
nautyextra_reroot_vpar (int *vpar, int n, int v)
{
  assert (n >= 0);
  assert (1 <= v && v <= n);
  int p = 0;  /* desired parent for v */
  while(v) {
    int new_v = vpar[v];
    vpar[v] = p; p = v; v = new_v;
  }
}

/* Change vpar[1..n] to reroot to a randomly chosen vertex.

   vpar should be a tree, not a forest.  Currently the re-rooting is to
   choose one vertex to be the root.  For a forest this means root of its
   component tree.

   Maybe could randomly reroot each component tree of a forest to a
   uniformly chosen vertex in that component.  Or maybe enough to randomly
   choose one vertex make it the root of its component.
*/
void
nautyextra_ranreroot_vpar (int *vpar, int n)
{
  if (n) {    /* where there are any vertices to set as root */
    int v; GETKRAN(n,v); v++;
    nautyextra_reroot_vpar(vpar,n, v);
  }
}


/* ------------------------------------------------------------------------- */
/* Make Various Graphs */

/* Set vpar[1..n] to a path of n vertices.
   vpar must have space for n+1 ints.
   Vertex 1 is the root, then successive children

       1 <-- 2 <-- 3 <-- 4      path-4
*/
void
nautyextra_make_path_vpar (int *vpar, int n)
{
  int i;
  for (i = 1; i <= n; i++)
    vpar[i] = i-1;
}

/* Set vpar[1..n] to a star of n vertices.
   vpar must have space for n+1 ints.

        1
       ^^^       star-4 (claw)
      / | \
     2  3  4

   Vertex 1 is the centre and the root, 2..n are the arms
*/
void
nautyextra_make_star_vpar (int *vpar, int n)
{
  if (n >= 1) {
    *++vpar = 0;  /* root */
    while (--n > 0)
      *++vpar = 1;
  }
}

/* cf SWHIBIT() which is the lowest 1-bit alone */
#define NAUTYEXTRA_MASK_LOWEST_ONE_AND_BELOW(n)   ((n) ^ ((n)-1))

/* ENHANCE-ME: use REMOVEHIBIT() new in nauty 2.7.x betas when ready */
#define NAUTYEXTRA_CLEAR_LOWEST_ONE(n)                  \
  ((n) & ~NAUTYEXTRA_MASK_LOWEST_ONE_AND_BELOW(n))

/* Set vpar[1..n] to a binomial tree of n vertices.
   The tree is defined by numbers m = 0 to n-1.  They are in vpar as 1 to n.
   Vertex v is number m=v-1.  Its parent is m with lowest 1-bit cleared to
   a 0-bit, which becomes vertex p = LOWEST_1_CLEARED(m) + 1.

      zero numbering      vpar vertex numbering
         __0___                 __1___
        /  |   \               /  |   \
       1   2    4             2   3    5
           |   / \                |   / \
           3  5   6               4  6   7
                  |                      |
                  7                      8
*/
void
nautyextra_make_binomial_tree_vpar (int *vpar, int n)
{
  DEBUG(printf("nautyextra_make_binomial_tree_vpar() n=%d\n", n));
  assert (n >= 0);
  if (n>0) {
    vpar[1] = 0; /* root */
    int v1;
    for (v1 = 2; v1 <= n; v1++) {
      int v0 = v1 - 1;
      int p0 = NAUTYEXTRA_CLEAR_LOWEST_ONE(v0);
      int p1 = p0 + 1;
      assert (0 <= p1 && p1 <= n);
      DEBUG(printf(" v1=%d p1=%d  v0=0x%X p0=0x%X\n", v1,p1, v0,p0));
      vpar[v1] = p1;
    }
  }
}

/* Set vpar[1..n] to a random tree n vertices.
   Parameter forest=0 gives a tree, or forest=1 is a forest.

   The vpar made is a uniform random choice over the labelled rooted trees
   or forests of n vertices.
*/
void
nautyextra_make_random_vpar (int *vpar, int n, int forest)
{
  /* This is a decode of the rooted labelled tree coding per

     .  Tamas Fleiner, "On Prufer Codes", Egervary Research Group Technical
     .  Report TR-2005-16.  https://web.cs.elte.hu/egres/tr/egres-05-16.pdf

     This decode has the attraction of needing no additional workspace, just
     some filling of vpar to identify vertices not yet seen.  Total work is
     n-1 random numbers and linear vpar filling and testing.

     The tree coding takes vertices in paths from root down to 1, then the
     highest unseen down to 2, then down to 3, etc, successively.  If a
     successive vertex v has already been seen because it was on a previous
     path, then there is nothing to go down to v and it is skipped.

     The tree code is the parent of each vertex taken in this sequence.  The
     first vertex in the sequence is the root so its parent is always 0.

     r, next_r are successive tree code values, made by GETKRAN().
     They are parents of vertices v, c, so vpar[v]=r and vpar[c]=next_r.
     When c is a child of v have its parent v = next_r and that allows
     vpar[v]=r to be set.

     When c is not a child of v will have next_r instead as some already
     seen vertex.  In that case v is instead the bottom of a run to the
     successive 1,2,etc.  "upto" is the next such bottom vertex.  upto may
     have been already seen (as noted above) and in that case search onwards
     for the lowest unseen.  Unseen vertices are flagged by vpar[]=-1.
  */

  DEBUG(printf("nautyextra_make_random_vpar() n=%d forest=%d\n", n, forest));
  assert (n >= 0);
  assert (forest==0 || forest==1);
  int limit = n + forest;
  int tree = 1 - forest;

  vpar[0] = 0;          /* 0 seen */
  if (n) {
    int i;
    for (i = 1; i <= n; i++)
      vpar[i] = -1;     /* all others unseen */

    int r = 0;
    int upto = 1;
    for (i = 1; i < n; i++) {
      int next_r;
#if NAUTYEXTRA_WANT_TEST_RANDOM
      next_r = nautyextra_test_random(limit);
#else
      GETKRAN(limit,next_r);
#endif
      next_r += tree;
      DEBUG(printf("at i=%d  r=%d next_r=%d\n", i, r, next_r));

      int v = next_r;
      assert (tree <= v && v <= n);  /* 1 to n for tree, 0 to n for forest */
      while (vpar[v] >= 0) {
        v = upto++;
        assert (1 <= v && v <= n);
      }

      DEBUG(printf("  store v=%d r=%d\n", v, r));
      assert (1 <= v && v <= n);
      vpar[v] = r;
      r = next_r;
    }

    /* Now have a single vpar[x]=-1, with x >= upto.
       Search upto ahead to find it.
    */
    assert (1 <= upto && upto <= n);
    while (vpar[upto] >= 0) {
      upto++;
      assert (1 <= upto && upto <= n);
    }
    vpar[upto] = r;
  }
}

/* ------------------------------------------------------------------------- */
/* Extensions and Modifications */

/* Subdivide the edges of vpar[1..n] by inserting a further s many vertices
   into each existing edge.
   Existing vertices remain as 1..n and new vertices are n+1 through to a
   new last vertex number n+s*(n-1).  The return is this new last vertex
   number, which is the new number of vertices.
   If s==0 then vpar is unchanged and the return is n.
   vpar[] must have space for the new vertices.
*/
int
nautyextra_subdivide_vpar (int *vpar, int n, int s)
{
  DEBUG(printf("nautyextra_subdivide_vpar() %d %d\n", n, s));
  assert (n >= 0);

  /* 1 <-- 2 <-- 3

     1 <-- 4 <-- 5 <-- 2 <-- 6 <-- 7 <-- 3
     v     d     e

     d = n+1 + (v-1)*s
     e = n+1 + (p-1)*s + (s-1) = n + p*s

     GP-Test  my(v=1,s=2,n=3); n+1 + (v-1)*s == 4
     GP-Test  my(p=1,s=2,n=3); n + p*s == 5
     GP-Test  my(s=2,n=3); n+s*(n-1) == 7
  */

  int last = n;  /* current last vertex of vpar */
  int v;
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p) {         /* insert into edge v-->p */
      int l;
      for (l = 0; l < s; l++) {
        vpar[++last] = p;
        DEBUG(printf(" new %d to %d\n", last, v));
        p = last;
      }
      vpar[v] = p;
    }
  }
  assert (last == n + s*nautyextra_num_edges_vpar(vpar,n));
  DEBUG(printf(" last %d\n", last));
  return last;
}


/* ------------------------------------------------------------------------- */
