#!/usr/bin/perl -w

# Copyright 2014, 2016 Kevin Ryde
#
# ps-to-png-pages.pl is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# ps-to-png-pages.pl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with ps-to-png-pages.pl.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.



#        make png          make r5dragon-png.html and its r5dragon-page-*.png
#                            (requires various pngcrop, optipng, etc)

# png: r5dragon-png.html
# r5dragon-png.html: r5dragon.ps r5dragon.pdf tools/ps-to-png-pages.pl Makefile
# 	perl tools/ps-to-png-pages.pl
# 	wc -c r5dragon-page-[0-9][0-9].png
	# rm -f r5dragon-png.html r5dragon-page-*.png
# There's too much mathematics and PGF pictures for the latex-to-text or
# latex-to-html tools to do any good, hence the r5dragon-png.html rendition
# in png images.
#
# clean:
#     r5dragon-tmp-cropped.pdf



# Usage: perl ps-to-png-pages.pl
#
# Read r5dragon.pdf and produce files which are png images of the pages and a
# html to look at them.
#
#      r5dragon-png.html
#      r5dragon-page-01.png
#      r5dragon-page-02.png
#      ...

use 5.006;
use strict;
use FindBin;
use File::Spec;
use File::Slurp 'slurp';
use Image::Size;
use List::Util 'max','min';
use IPC::Run;
use POSIX ();

# uncomment this to run the ### lines
# use Smart::Comments;

my $progname = $FindBin::Script; # basename part
print "progname '$progname'\n";


my $AlphaBits = 4;   # 1,2,4
my $html_filename = 'r5dragon-png.html';
# my $software = "Generated by $progname";

# PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
my $time_str = POSIX::strftime (STRFTIME_FORMAT_RFC822, localtime(time));

my $script_filename = File::Spec->catfile($FindBin::Bin, $FindBin::Script);
$script_filename = File::Spec->rel2abs($script_filename);
print "script $script_filename\n";

my $source_dir = File::Spec->catdir ($FindBin::Bin, File::Spec->updir);
$source_dir = File::Spec->rel2abs($source_dir);
print "source dir $source_dir\n";

chdir $source_dir or die $!;

my $tex_filename = 'r5dragon.tex';
my $pdf_filename = 'r5dragon.pdf';
my $ps_filename = 'r5dragon.ps';

my $version;
IPC::Run::run(['make','--no-print-directory','version'],
              '<', File::Spec->devnull,
              '>', \$version,
              '2>', File::Spec->devnull) or die;
$version =~ s/\s+$//;
$version =~ /^\d+$/ or die "Bad version number: '$version'";
print "version $version\n";

my $date_yyyy_mm_dd;
{
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
      $atime,$mtime,$ctime,$blksize,$blocks)
    = stat($tex_filename)
      or die "Cannot stat $tex_filename: $!";
  $date_yyyy_mm_dd = POSIX::strftime ('%Y-%m-%d', localtime(time));
}


#------------------------------------------------------------------------------
# title, from r5dragon.tex

my $title;
{

  my $tex = File::Slurp::slurp($tex_filename);
  $tex =~ /^\\title\{([^}]+)\}/m
    or die "\\title not found";
  $title = $1;
}
print "title:     $title\n";

#------------------------------------------------------------------------------
# copyright, from self script

my $copyright_str = do {
  my $slurp = slurp($script_filename);
  $slurp =~ /^# (Copyright.*)/m or die;
  $1;
};
print "copyright: $copyright_str\n";

#------------------------------------------------------------------------------
# remove old png files

foreach my $png_filename (glob "r5dragon-page-*.png") {
  unlink $png_filename or die;
}

#------------------------------------------------------------------------------
# cropped pdf

# $filename is a pdf or ps.
# Return ($x1,$y1, $x2,$y2) which are bounding box corners maximum and
# minimum of the bounding boxes of all the pages.  So the bounding box
# returned suffices for all pages.
#
sub pdf_bbox_maximum {
  my ($filename) = @_;

  my $bbox_out;
  IPC::Run::run(['gs','-dNOPAUSE','-sDEVICE=bbox',$filename],
                '<', File::Spec->devnull,
                '>', \$bbox_out,
                '2>&1')
      or die "ghostscript error:\n$bbox_out\n";
  ### $bbox_out

  my @bboxes;
  while ($bbox_out =~ /%%BoundingBox: (\d+) (\d+) (\d+) (\d+)/g) {
    print "bbox $1,$2 $3,$4\n";
    push @bboxes, [$1,$2,$3,$4];
  }
  return (min(map {$_->[0]} @bboxes),  # x1
          min(map {$_->[1]} @bboxes),  # y1
          max(map {$_->[2]} @bboxes),  # x2
          max(map {$_->[3]} @bboxes)); # y2
}

# stretch bounding box for some margin
sub bbox_stretch {
  my ($x1,$y1, $x2,$y2) = @_;
  my $hstretch = int(($x2-$x1) * .04);
  my $vstretch = int(($y2-$y1) * .04);
  $hstretch = min($hstretch,$x1);
  $vstretch = min($vstretch,$y1);
  return ($x1 - $hstretch,
          $y1 - $vstretch,
          $x2 + $hstretch,
          $y2 + $vstretch);
}

my ($x1,$y1, $x2,$y2) = pdf_bbox_maximum($ps_filename);  # $pdf_filename
print "max  $x1,$y1 $x2,$y2\n";

($x1,$y1, $x2,$y2) = bbox_stretch ($x1,$y1, $x2,$y2);
print "stretched $x1,$y1 $x2,$y2\n";

my $pdfcrop_filename = 'r5dragon-tmp-cropped.pdf';
IPC::Run::run(['pdfcrop',
               # ,'--verbose'
               '--bbox',"$x1 $y1 $x2 $y2",
               $pdf_filename, $pdfcrop_filename])
  or die;


#------------------------------------------------------------------------------
# pngs by ghostscript

print "gs\n";
IPC::Run::run(['gs',
               '-q', # quiet, cf -dQUIET
               '-dNOPAUSE',
               '-r100',
               '-sDEVICE=pnggray',
               "-dTextAlphaBits=$AlphaBits",     # gray scale edge blurring
               "-dGraphicsAlphaBits=$AlphaBits",
               # '-dGrayValues=16', # only ppm, not png
               '-sOutputFile=r5dragon-page-%02d.png', $pdfcrop_filename],
              '<', File::Spec->devnull)
  or die;
unlink($pdfcrop_filename) or die;
print "\n";

my @png_filenames = (sort glob("r5dragon-page-[0-9][0-9].png"))
  or die "oops, no png image files";
my $max_pagenum = scalar(@png_filenames) || die;
my $total_size = int(List::Util::sum(map {-s} @png_filenames) / 1000);
print "max pagenum $max_pagenum total size ${total_size}k\n";

#------------------------------------------------------------------------------
# pngs text

foreach my $png_filename (@png_filenames) {
  ### $png_filename
  my ($pagenum02) = ($png_filename =~ /(\d+)/);
  my $pagenum = $pagenum02 + 0;

  foreach my $elem
    ([ Author     => 'Kevin Ryde' ],
     [ Title      => "Iterations of the R5 Dragon Curve - page $pagenum"],
     [ Version    => "Draft $version" ],
     [ Copyright  => $copyright_str ],
     [ 'Creation Time' => $time_str ],
     [ Homepage        => 'http://user42.tuxfamily.org/r5dragon/index.html' ],
    ) {
    my ($keyword, $value) = @$elem;
    IPC::Run::run(['pngtextadd',
                   '--keyword',$keyword,
                   '--text',$value,
                   $png_filename]) or die;
  }
}
# pngtextadd --keyword=Software --text='$software' $png_filename &&

#------------------------------------------------------------------------------
# optipng
#
# cf pngcrush can add tEXt too
#    pngquant,pngnq 32-bit RGB to 8-bit palette by dithering

{
  IPC::Run::run(['optipng','-quiet',
                 '-o2',   # is -zc9 and a few trials
                 @png_filenames])
      or die;
  $total_size = int(List::Util::sum(map {-s} @png_filenames) / 1000);
  print "optipng reduced to total size ${total_size}k\n";
}

#------------------------------------------------------------------------------
# r5dragon-png.html

{
  open my $fh, '>', $html_filename or die;
  print $fh <<"HERE";
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">
<html lang="en">
<head>
<title>$title - PNG</title>
<meta name="description"
      content="A paper on the R5 dragon curve - PNG page images.">
<meta name="Author" content="Kevin Ryde">
<meta name="Generator" content="$progname">
<link rel="icon" href="favicon.png" type="image/png"/>
</head>

<body itemscope itemtype="http://schema.org/WebPage">
<meta itemprop="inLanguage" content="English"/>
<meta itemprop="dateCreated" content="$date_yyyy_mm_dd"/>
<meta itemprop="datePublished" content="$date_yyyy_mm_dd"/>
<meta itemprop="dateModified" content="$date_yyyy_mm_dd"/>

<h1 align="center">Iterations of the R5 Dragon Curve</h1>

<p>
Draft $version

<p>
Page:
HERE
  foreach my $pagenum (1 .. $max_pagenum) {
    printf $fh "<a href=\"#page%02d\">%02d</a>\n", $pagenum, $pagenum;
  }
  print $fh <<"HERE";
<br>
Total images size ${total_size}k.

<p>
(<a href="http://user42.tuxfamily.org/r5dragon/index.html"
    rel="home">Back to the r5dragon paper page.</a>)
</p>
HERE

  foreach my $pagenum (1 .. $max_pagenum) {
    my $pagenum02 = sprintf '%02d', $pagenum;
    my $png_filename = $png_filenames[$pagenum-1];

    my ($width,$height) = Image::Size::imgsize($png_filename);
    my $size_kbytes = int((-s $png_filename) / 1000);
    print "$png_filename ${width}x${height} ${size_kbytes}k\n";

    print $fh <<"HERE";
<hr width="100%">
<p>
<a name="page$pagenum02">Page $pagenum</a>
<br>
<img src="$png_filename" width=$width height=$height
     alt="Page $pagenum02, $size_kbytes kbytes">
</p>

HERE
  }

  my $schema_copyright_str = $copyright_str;
  $schema_copyright_str =~ s{(\d )(\D.*)}
                            {$1<span itemprop="copyrightHolder">$2</span>};
  print $fh <<"HERE";
<hr width="100%">
<p>
<small>This page $schema_copyright_str.</small>
<p>
<small>(<a href="http://user42.tuxfamily.org/r5dragon/index.html"
           rel="home">Back to the R5 dragon paper page.</a>)</small>
</p>
</body>
</html>
HERE

  close $fh or die;
}

#------------------------------------------------------------------------------
# weblint

if (IPC::Run::run(['weblint',$html_filename])) {
  print "weblint ok\n";
} else {
  die;
}

#------------------------------------------------------------------------------

exit 0;
